# This program is public domain
"""
Load a NeXus file into a reflectometry data structure.
"""
import os

import numpy as np

from reductus.dataflow.lib import unit
from reductus.dataflow.lib import iso8601
from reductus.dataflow.lib import h5_open
from reductus.dataflow.lib.strings import _s, _b

from .refldata import ReflData
from .resolution import FWHM2sigma

TRAJECTORY_INTENTS = {
    'SPEC': 'specular',
    'SLIT': 'intensity',
    'BGP': 'background+',
    'BGM': 'background-',
    'ROCK': 'rock sample',
}

WAVELENGTH = 4.
WAVELENGTH_DISPERSION = 0.015

def data_as(group, fieldname, units, rep=None, NA=None, dtype=None):
    """
    Return value of field in the desired units.
    """
    if fieldname not in group:
        return NA
    field = group[fieldname]
    units_in = _s(field.attrs.get('units', ''))
    converter = unit.Converter(units_in)
    value = converter(field[()], units)
    if dtype is not None:
        value = np.asarray(value, dtype=dtype)
    if rep is not None:
        if np.isscalar(value) or len(value) == 1:
            return np.repeat(value, rep, axis=0)
        elif len(value) == rep:
            return value
        else:
            raise ValueError("field %r does not match counts in %r"
                             %(field.name, field.file.filename))
    else:
        return value


def str_data(group, field, default=''):
    """
    Retrieve value of field as a string, with default if field is missing.
    """
    if field in group:
        data = group[field][0]
        if data.ndim > 0:
            value = [_s(v) for v in data]
        else:
            value = _s(data)
        return value
    else:
        return default


def list_data(group, field):
    """
    Retrieve value of field as a list of strings, or []
    """
    if field in group:
        return [_s(s) for s in group[field]]
    else:
        return []


def nxfind(group, nxclass):
    """
    Iterate over the entries of type *nxclass* in the hdf5 *group*.
    """
    for entry in group.values():
        if nxclass == _s(entry.attrs.get('NX_class', None)):
            yield entry


def load_metadata(filename, file_obj=None):
    """
    Load the summary info for all entries in a NeXus file.
    """
    return load_nexus_entries(filename, file_obj=file_obj,
                              meta_only=True, entry_loader=NCNRNeXusRefl)


def load_entries(filename, file_obj=None, entries=None):
    return load_nexus_entries(filename, file_obj=file_obj, entries=entries,
                              meta_only=False, entry_loader=NCNRNeXusRefl)


def load_nexus_entries(filename, file_obj=None, entries=None,
                       meta_only=False, entry_loader=None):
    """
    Load the summary info for all entries in a NeXus file.
    """
    handle = h5_open.h5_open_zip(filename, file_obj)
    measurements = []
    for name, entry in handle.items():
        if entries is not None and name not in entries:
            continue
        if _s(entry.attrs.get('NX_class', None)) == 'NXentry':
            data = entry_loader(entry, name, filename)
            if not meta_only:
                data.load(entry)
            measurements.append(data)
    if file_obj is None:
        handle.close()
    return measurements


def nexus_common(self, entry, entryname, filename):
    #print(entry['instrument'].values())
    das = entry['DAS_logs']
    self.entry = entryname
    self.path = os.path.abspath(filename)
    self.name = str_data(das, 'trajectoryData/fileName', 'unknown')
    if 'trajectoryData/fileNum' in das:
        self.filenumber = das['trajectoryData/fileNum'][0]
    else:
        # fall back to randomly generated filenum
        from random import randint
        self.filenumber = -randint(10**9, (10**10) - 1)

    #self.date = iso8601.parse_date(entry['start_time'][0].decode('utf-8'))
    self.date = iso8601.parse_date(str_data(entry, 'start_time'))
    self.description = str_data(entry, 'experiment_description')
    self.instrument = str_data(entry, 'instrument/name')

    # Determine the number of points in the scan.
    # TODO: Reliable way to determine scan length.
    if 'trajectory/liveScanLength' in entry:
        # New files should have num points in trajectory/liveScanLength ...
        n = entry['trajectory/liveScanLength'][()]
    else:
        # Guess length by looking at the counter fields
        # Prefer to not load the entire counter at this point, especially since
        # we don't know where it is.
        n = das['counter/liveROI'].shape[0]
        if n == 1:
            n = das['counter/liveMonitor'].shape[0]
        if n == 1:
            n = das['counter/liveTime'].shape[0]
    self.points = n

    monitor_device = entry.get('control/monitor', {})
    self.monitor.deadtime = data_as(monitor_device, 'dead_time', 'us')
    self.monitor.deadtime_error = data_as(monitor_device, 'dead_time_error', 'us')
    base = str_data(das, 'counter/countAgainst').lower()
    # NICE stores TIME, MONITOR, ROI, TIME_MONITOR, TIME_ROI, etc.
    if "monitor" in base:
        base = "monitor"
    elif "time" in base:
        base = "time"
    elif "roi" in base:
        base = "roi"
    else:
        base = "none"

    self.monitor.time_step = 0.001  # assume 1 ms accuracy on reported clock
    self.monitor.counts = data_as(das, 'counter/liveMonitor', '', rep=n, dtype='d')
    self.monitor.counts_variance = self.monitor.counts.copy()
    self.monitor.count_time = data_as(das, 'counter/liveTime', 's', rep=n)
    self.monitor.roi_counts = data_as(das, 'counter/liveROI', '', rep=n, dtype='d')
    self.monitor.roi_variance = self.monitor.roi_counts.copy()
    self.monitor.roi_variance = self.monitor.roi_counts.copy()
    self.monitor.source_power = data_as(das,
        'reactorPower/reactorPowerThermal/average_value', 'MW', rep=n, dtype='d')
    self.monitor.source_power_variance = data_as(das, 
        'reactorPower/reactorPowerThermal/average_value_error', 'MW', rep=n, dtype='d')
    self.monitor.source_power_units = "MW"

    # NG7 monitor saturation is stored in control/countrate_correction
    saturation_device = entry.get('control/countrate_correction', None)
    if saturation_device is not None:
        rate = data_as(saturation_device, 'measured_rate', '')
        correction = data_as(saturation_device, 'correction', '')
        self.monitor.saturation = np.vstack((rate, 1./correction))

    # CRUFT: old candor files don't define NXsample
    self.sample.name = str_data(entry, 'sample/name', default=None)
    self.sample.description = str_data(entry, 'sample/description')
    if self.sample.name is None:
        self.sample.name = str_data(entry, 'DAS_logs/sample/name')
        self.sample.description = str_data(entry, 'DAS_logs/sample/description')

    # TODO: stop trying to guess DOI
    if 'DOI' in entry:
        URI = _s(entry['DOI'])
    else:
        # See: dataflow.modules.doi_resolve for helpers.
        #NCNR_DOI = "10.18434/T4201B"
        NCNR_DOI = "https://ncnr.nist.gov/pub/ncnrdata"
        LOCATION = {'pbr':'ngd', 'magik':'cgd', 'ng7r':'ng7', 'candor':'cdr'}
        nice_instrument = str_data(das, 'experiment/instrument').lower()
        instrument = LOCATION.get(nice_instrument, nice_instrument)
        year, month = self.date.year, self.date.month
        cycle = "%4d%02d"%(year, month)
        experiment = str_data(entry, 'experiment_identifier')
        filename = os.path.basename(self.path)
        URI = "/".join((NCNR_DOI, instrument, cycle, experiment, "data", filename))
    self.uri = URI

    self.scan_value = []
    self.scan_units = []
    self.scan_label = []
    if 'trajectory/scannedVariables' in das:
        scanned_variables = list_data(das, 'trajectory/scannedVariables')
        # Just in case the scanned variables is a string with
        # elements separated by new lines...
        if len(scanned_variables) == 1:
            scanned_variables = scanned_variables[0].split('\n')
        # TODO: exclude count fields from scanned variables
        #scanned_variables = [s for s in scanned_variables
        #                     if not s.startswith("areaDetector")]
        for node_id in scanned_variables:
            path = node_id.replace('.', '/')
            try:
                field = das[path]
            except KeyError:
                # Note: Suppressing this message because it makes the
                # regression tests noisy.  Older versions of the SelectFields
                # filter on the datawriter were not stripping the fields from
                # control/scanned variables lists, but newer ones are.
                # TODO: reenable test for missing scan fields
                #print(">>> could not read scanned %s for %s"
                #      % (node_id, os.path.basename(self.path)))
                continue
            try:
                scan_value = data_as(das, path, '', rep=n)
                scan_units = _s(field.attrs.get('units', ''))
                scan_label = _s(field.attrs.get('label', node_id))
            except Exception as exc:
                print(">>> unexpected error %s reading %s for %s"
                      % (str(exc), node_id, os.path.basename(self.path)))
                continue
            # check if numeric:
            if scan_value.dtype.kind in ["f", "u", "i"]:
                self.scan_value.append(scan_value)
                self.scan_units.append(scan_units)
                self.scan_label.append(scan_label)

    # TODO: magnetic field
    if 'temp' in das:
        if 'temp/primaryControlLoop' in das:
            temp_controller = das['temp/primaryControlLoop'][0]
            setpoint_field = 'temp/setpoint_%d' % (temp_controller,)
            self.sample.temp_setpoint = data_as(das, setpoint_field, 'K')[0]
        temp_values = data_as(das, 'temp/primaryNode/value', 'K')
        temp_shape = temp_values.shape[0] if temp_values.shape[0] else 1.0
        # only include one significant figure for temperatures.
        self.sample.temp_avg = round(np.sum(temp_values)/temp_shape, 1)


def get_pol(das, pol):
    if pol in das:
        direction = str_data(das, pol+'/direction')
        if direction == 'UP':
            result = '+'
        elif direction == 'DOWN':
            result = '-'
        elif direction in ('', 'BEAM_OUT', 'UNPOLARIZED'):
            result = ''
        else:
            raise ValueError("Don't understand DAS_logs/%s/direction=%r"
                             %(pol, direction))
    else:
        result = ''
    return result


class NCNRNeXusRefl(ReflData):
    """
    NeXus reflectometry entry.

    See :class:`ReflData` for details.
    """
    format = "NeXus"
    probe = "neutrons"

    def __init__(self, entry, entryname, filename):
        super(NCNRNeXusRefl, self).__init__()
        nexus_common(self, entry, entryname, filename)

    def load(self, entry):
        #print(entry['instrument'].values())
        das = entry['DAS_logs']
        n = self.points
        raw_intent = str_data(das, 'trajectoryData/_scanType')
        if raw_intent in TRAJECTORY_INTENTS:
            self.intent = TRAJECTORY_INTENTS[raw_intent]

        # Polarizers
        self.polarization = (
            get_pol(das, 'frontPolarization')
            + get_pol(das, 'backPolarization')
        )

        # Monochromator
        self.monochromator.wavelength = data_as(entry, 'instrument/monochromator/wavelength', 'Ang', rep=n)
        self.monochromator.wavelength_resolution = data_as(entry, 'instrument/monochromator/wavelength_error', 'Ang', rep=n)
        if self.monochromator.wavelength is None:
            # for old NG7 data, wavelength is in DAS_logs
            self.monochromator.wavelength = data_as(das, 'wavelength/wavelength', 'Ang', rep=n)
        if self.monochromator.wavelength is None:
            self.warn("Wavelength is missing; using {WAVELENGTH} A".format(WAVELENGTH=WAVELENGTH))
            self.monochromator.wavelength = WAVELENGTH
        if self.monochromator.wavelength_resolution is None:
            self.warn("Wavelength resolution is missing; using %.1f%% dL/L FWHM"
                      % (100*WAVELENGTH_DISPERSION,))
            self.monochromator.wavelength_resolution = \
                FWHM2sigma(WAVELENGTH_DISPERSION*self.monochromator.wavelength)

        # Slits
        self.slit1.distance = data_as(entry, 'instrument/presample_slit1/distance', 'mm')
        self.slit2.distance = data_as(entry, 'instrument/presample_slit2/distance', 'mm')
        self.slit3.distance = data_as(entry, 'instrument/predetector_slit1/distance', 'mm')
        self.slit4.distance = data_as(entry, 'instrument/predetector_slit2/distance', 'mm')
        if self.slit1.distance is None:
            self.warn("Slit 1 distance is missing; using 2 m")
            self.slit1.distance = -2000
        if self.slit2.distance is None:
            self.warn("Slit 2 distance is missing; using 1 m")
            self.slit2.distance = -1000

        for k, slit in enumerate([self.slit1, self.slit2, self.slit3, self.slit4]):
            x = 'slitAperture%d/softPosition'%(k+1)
            x_target = 'slitAperture%d/desiredSoftPosition'%(k+1)
            slit.x = data_as(das, x, 'mm', rep=n, dtype='d')
            slit.x_target = data_as(das, x_target, 'mm', rep=n, dtype='d')
            y = 'vertSlitAperture%d/softPosition'%(k+1)
            y_target = 'vertSlitAperture%d/desiredSoftPosition'%(k+1)
            slit.y = data_as(das, y, 'mm', rep=n, dtype='d')
            slit.y_target = data_as(das, y_target, 'mm', rep=n, dtype='d')

        # Detector
        self.detector.wavelength = self.monochromator.wavelength
        self.detector.wavelength_resolution = self.monochromator.wavelength_resolution
        self.detector.deadtime = data_as(entry, 'instrument/single_detector/dead_time', 'us')
        self.detector.deadtime_error = data_as(entry, 'instrument/single_detector/dead_time_error', 'us')
        self.detector.distance = data_as(entry, 'instrument/detector/distance', 'mm')
        self.detector.rotation = data_as(entry, 'instrument/detector/rotation', 'degree')

        # Counts
        self.detector.counts = data_as(das, 'counter/liveROI', '', dtype='d')
        self.detector.counts_variance = self.detector.counts.copy()
        self.detector.dims = self.detector.counts.shape[1:]

        # Angles
        if 'sampleAngle' in das:
            # selects MAGIK or PBR, which have sample and detector angle
            self.sample.angle_x = data_as(das, 'sampleAngle/softPosition', 'degree', rep=n)
            self.detector.angle_x = data_as(das, 'detectorAngle/softPosition', 'degree', rep=n)
            self.sample.angle_x_target = data_as(das, 'sampleAngle/desiredSoftPosition', 'degree', rep=n)
            self.detector.angle_x_target = data_as(das, 'detectorAngle/desiredSoftPosition', 'degree', rep=n)
        elif 'q' in das:
            # selects NG7R which has only q device (qz) and sampleTilt
            # Ignore sampleTilt for now since it is arbitrary.  NG7 is not
            # using zeros for the sampleTilt motor in a predictable way.
            tilt = 0.
            #tilt = data_as(das, 'sampleTilt/softPosition', 'degree', rep=n)
            theta = data_as(das, 'q/thetaIncident', 'degree', rep=n)
            if theta is not None:
                self.sample.angle_x = theta + tilt
                self.detector.angle_x = 2*theta
            # NaN if not defined
            tilt_target = 0.
            #tilt_target = data_as(das, 'sampleTilt/desiredSoftPosition', 'degree', rep=n)
            # Note: q/desiredThetaIncident is not available on any instruments
            # so the following always returns None.
            theta_target = data_as(das, 'q/desiredThetaIncident', 'degree', rep=n)
            if theta_target is not None:
                self.sample.angle_x_target = theta_target + tilt_target
                self.detector.angle_x_target = 2*theta_target
        else:
            raise ValueError("Unknown sample angle in file")
        self.Qz_target = data_as(das, 'q/z', '', rep=n)
        if self.Qz_target is None:
            self.Qz_target = data_as(das, 'trajectoryData/_q', '', rep=n)
        # TODO: use background_offset if it is defined
        #if 'trajectoryData/_theta_offset' in das:
        #    self.background_offset = 'theta'

    def _load_slits(self, instrument):
        """
        Slit names have not been standardized.  Instead sort the
        NXaperature components by distance and assign them according
        to serial order, negative aperatures first and positive second.
        """
        raise NotImplementedError("hardcode slit names for now")
        slits = list(nxfind(instrument, 'NXaperature'))
        # Note: only supports to aperatures before and after.
        # Assumes x and y aperatures are coupled in the same
        # component.  This will likely be wrong for some instruments,
        # but we won't deal with that until we have real NeXus files
        # to support.
        # Assume the file writer was sane and specified all slit
        # distances in the same units so that sorting is simple.
        # Currently only slit distance is recorded, not slit opening.
        slits.sort(lambda a,b: -1 if a.distance < b.distance else 1)
        index = 0
        for slit in slits:
            d = slit.distance.value('meters')
            if d <= 0:
                # process first two slits only
                if index == 0:
                    self.slit1.distance = d
                    index += 1
                elif index == 1:
                    self.slit2.distance = d
                    index += 1
            elif d > 0:
                # skip leading slits
                if index < 2: index = 2
                if index == 2:
                    self.slit3.distance = d
                    index += 1
                elif index == 3:
                    self.slit4.distance = d
                    index += 1


def demo(loader=load_entries):
    import sys
    from .load import setup_fetch, fetch_uri
    if len(sys.argv) == 1:
        print("usage: python -m reflred.nexusref file...")
        sys.exit(1)
    setup_fetch()
    for uri in sys.argv[1:]:
        try:
            entries = fetch_uri(uri, loader=loader)
        except Exception as exc:
            print("**** "+str(exc)+" **** while reading "+uri)
            raise
            continue

        # print the first entry
        print(entries[0])

        # plot all the entries
        import pylab
        pylab.figure()
        for entry in entries:
            entry.plot()
        pylab.legend()
        pylab.show()

if __name__ == "__main__":
    demo()



# -*- coding: UTF-8 -*-
import gzip
from collections import OrderedDict
import datetime
import numpy as np
import json
import io
import sys

IS_PY3 = sys.version_info[0] >= 3

def _b(s):
    if IS_PY3:
        return s.encode('utf-8')
    else:
        return s

from reductus.dataflow.lib import octave
from reductus.dataflow.lib.exporters import exports_HDF5, exports_text

class RawData(object):
    def __init__(self, name, data):
        histo_array = data.pop("histodata")
        self.histodata = histo_array
        #self.histodata = Uncertainty(histo_array, histo_array)
        self.metadata = {
            "name": name,
            "entry": "entry"
        }
        self.metadata.update(data)
        self.name = name

    def todict(self):
        return _toDictItem(self.metadata)

    def get_plottable(self):
        return {"entry": "entry", "type": "metadata", "values": _toDictItem(self.metadata)}

    def get_metadata(self):
        return _toDictItem(self.metadata)

class EfTwoThetaData(object):
    def __init__(self, name, data, ef=None, twotheta=None, metadata=None):
        self.metadata = metadata
        self.data = data
        self.ef = ef
        self.twotheta = twotheta
        self.name = name
        self.xaxis = {
            "label": u"Ef [meV]",
            "values": self.ef,
            "min": self.ef.min(),
            "max": self.ef.max(),
            "dim": self.ef.shape[0]
        }
        self.yaxis = {
            "label": "Detector Angle [degrees]",
            "values": self.twotheta,
            "min": self.twotheta.min(),
            "max": self.twotheta.max(),
            "dim": self.twotheta.shape[0]
        }

    def get_plottable(self):
        output = {
            "title": self.name,
            "dims": {
                "ymin": self.yaxis["min"],
                "ymax": self.yaxis["max"],
                "ydim": self.yaxis["dim"],
                "xmin": self.xaxis["min"],
                "xmax": self.xaxis["max"],
                "xdim": self.xaxis["dim"],
                "zmin": self.data.min(),
                "zmax": self.data.max()
            },
            "type": "2d",
            "xlabel": self.xaxis["label"],
            "ylabel": self.yaxis["label"],
            "z": [self.data.flatten().tolist()]
        }
        return output
    
    def get_metadata(self):
        return _toDictItem(self.metadata)

class EQData(object):
    def __init__(self, name, data, metadata):
        self.metadata = metadata
        self.data = data
        self.name = name
        self.xaxis = {
            "label": u"|Q| [A⁻¹]",
            "values": np.linspace(metadata["Q_min"], metadata["Q_max"], data.shape[0]),
            "min": metadata["Q_min"],
            "max": metadata["Q_max"],
            "dim": data.shape[0]
        }
        self.yaxis = {
            "label": "Ei-Ef [meV]",
            "values": np.linspace(-metadata["Ei"], metadata["Ei"], data.shape[1]),
            "min": -metadata["Ei"],
            "max": metadata["Ei"],
            "dim": data.shape[1]
        }

    def get_plottable(self):
        Ei = self.metadata['Ei']
        Q_max = self.metadata['Q_max']
        EQ_data = self.data
        output = {
            "title": self.name,
            "dims": {
                "ymin": self.yaxis["min"],
                "ymax": self.yaxis["max"],
                "ydim": self.yaxis["dim"],
                "xmin": self.xaxis["min"],
                "xmax": self.xaxis["max"],
                "xdim": self.xaxis["dim"],
                "zmin": self.data.min(),
                "zmax": self.data.max()
            },
            "type": "2d",
            "xlabel": self.xaxis["label"],
            "ylabel": self.yaxis["label"],
            "z": [self.data.flatten().tolist()]
        }
        return output

    def get_metadata(self):
        return _toDictItem(self.metadata) 

class DCS1dData(object):
    properties = ['x', 'v', 'dx', 'dv', 'xlabel', 'vlabel', 'xunits', 'vunits', 'metadata']

    def __init__(self, x, v, dx=0, dv=0, xlabel="", vlabel="", xunits="", vunits="", metadata=None):
        self.x = x
        self.v = v
        self.dx = dx
        self.dv = dv
        self.xlabel = xlabel
        self.vlabel = vlabel
        self.xunits = xunits
        self.vunits = vunits
        self.metadata = metadata if metadata is not None else {}

    def to_dict(self):
        props = dict([(p, getattr(self, p, None)) for p in self.properties])
        return _toDictItem(props)

    def get_plottable(self):
        label = "%s: %s" % (self.metadata['name'], self.metadata['entry'])
        xdata = self.x.tolist()
        ydata = self.v.tolist()
        yerr = self.dv.tolist()
        data = [[x, y, {"yupper": y+dy, "ylower": y-dy, "xupper": x, "xlower": x}] for x,y,dy in zip(xdata, ydata, yerr)]
        plottable = {
            "type": "1d",
            "title": self.metadata.get("name", "DCS 1d data"),
            "options": {
                "axes": {
                    "xaxis": {"label": self.xlabel},
                    "yaxis": {"label": self.vlabel}
                },
                "series": [{"label": label}]
            },
            "data": [data]
        }
        return plottable

    def get_metadata(self):
        return self.to_dict()

    @exports_text(name="column")
    def to_column_text(self):
        with io.BytesIO() as fid:
            #fid.write(_b("# %s\n" % json.dumps(_toDictItem(self.metadata)).strip("{}")))
            metadata = {"name": self.metadata["name"]}
            fid.write(_b("# %s\n" % json.dumps(metadata).strip("{}")))
            columns = {"columns": [self.xlabel, self.vlabel, "uncertainty", "resolution"]}
            units = {"units": [self.xunits, self.vunits, self.vunits, self.xunits]}
            fid.write(_b("# %s\n" % json.dumps(columns).strip("{}")))
            fid.write(_b("# %s\n" % json.dumps(units).strip("{}")))
            np.savetxt(fid, np.vstack([self.x, self.v, self.dv, self.dx]).T, fmt="%.10e")
            fid.seek(0)
            value = fid.read()

        return {
            "name": self.metadata.get("name", "default_name"),
            "entry": self.metadata.get("entry", "default_entry"),
            "file_suffix": ".dcs.dat",
            "value": value.decode(),
        }

class Parameters(dict):
    def get_metadata(self):
        return _toDictItem(self)

    def get_plottable(self):
        return {"entry": "entry", "type": "metadata", "values": _toDictItem(self)}

def readDCS(name, fid):
    gzf = gzip.GzipFile(fileobj=fid)
    data = octave.read_octave_binary(gzf)
    return RawData(name, data)

def _toDictItem(obj):
    if isinstance(obj, np.integer):
        obj = int(obj)
    elif isinstance(obj, np.floating):
        obj = float(obj)
    elif isinstance(obj, np.ndarray):
        obj = obj.tolist()
    elif isinstance(obj, datetime.datetime):
        obj = [obj.year, obj.month, obj.day, obj.hour, obj.minute, obj.second]
    elif isinstance(obj, list):
        obj = [_toDictItem(a) for a in obj]
    elif isinstance(obj, dict):
        obj = dict([(k, _toDictItem(v)) for k, v in obj.items()])
    return obj


import copy
from xml.etree import ElementTree

import numpy as np

def load(file_or_filename):
    tree = ElementTree.parse(file_or_filename)
    root = tree.getroot()
    ns = {'': root.attrib[root.keys()[0]].split()[0]}

    context = {}
    sample_id = root.find("./sample/id", ns).text
    context['sample_id'] = "" if sample_id is None else sample_id
    
    sample_name = getattr(root.find("./sample/name", ns), "text", "")
    context['sample_name'] = "" if sample_name is not None else sample_name

    measurement_nodes = root.findall("./xrdMeasurement[@measurementType='Scan']", ns)
    entries = []
    for node in measurement_nodes:
        entries.extend(parse_measurement_node(node, ns, context))
    return entries
    
def parse_measurement_node(node, ns, context):
    comments = node.findall('./comment/entry', ns)
    context['comment'] = "" if comments is None else "\n".join([n.text for n in comments if n.text is not None])
    wavelength = node.find('./usedWavelength', ns)
    context['kAlpha1'] = float(wavelength.find('./kAlpha1', ns).text)
    context['kAlpha2'] = float(wavelength.find('./kAlpha2', ns).text)
    kb = wavelength.find('./kBeta', ns).text
    context['ratioKAlpha2KAlpha1'] = float(wavelength.find('./ratioKAlpha2KAlpha1', ns).text)
    # output['wavelength'] = (ka1 + kratio * ka2) / (kratio + 1.0)

    context['incidentRadius'] = float(node.find("./incidentBeamPath/radius", ns).text)
    context['sourceLineWidth'] = float(node.find("./incidentBeamPath/xRayTube/focus/width", ns).text)

    divergence_slit = node.find("./incidentBeamPath/divergenceSlit", ns)
    if divergence_slit is not None:
        context['divergenceSlitDistance'] = float(divergence_slit.find("./distanceToSample", ns).text)
        context['divergenceSlitHeight'] = float(divergence_slit.find("./height", ns).text)

    context['diffractedRadius'] = float(node.find("./diffractedBeamPath/radius", ns).text)
    
    detector = node.find("./diffractedBeamPath/detector", ns)
    detectorTypeKey = next(k for k in detector.attrib if k.endswith('type'))
    detectorType = detector.attrib[detectorTypeKey]
    
    receivingSlitHeight = None
    receiving_slit = node.find("./diffractedBeamPath/receivingSlit", ns)
    if receiving_slit is not None:
        receivingSlitHeight = float(receiving_slit.find("./height", ns).text)

    active_length = node.find("./diffractedBeamPath/detector/activeLength", ns)
    if active_length is not None:
        receivingSlitHeight = float(active_length.text)

    context['receivingSlitHeight'] = receivingSlitHeight
    

    scan_nodes = node.findall('./scan', ns)
    output = []
    for s in scan_nodes:
        scan_context = copy.deepcopy(context)
        scan_context.update(parse_scan_node(s, ns))
        output.append(scan_context)
    return output

def get_positions(node, axis_name, ns, numpoints):
    common = node.find("./dataPoints/positions[@axis='{}']/commonPosition".format(axis_name), ns)
    positions_list = node.find("./dataPoints/positions[@axis='{}']/listPositions".format(axis_name), ns)

    if common is not None:
        return np.ones((numpoints,), dtype=float) * float(common.text)
    
    elif positions_list is not None:
        # does it need truncation?  is it always the same length as counts?
        return np.fromstring(positions_list.text, dtype=float, sep=" ")[:numpoints]
    
    else:
        start = float(node.find("./dataPoints/positions[@axis='{}']/startPosition".format(axis_name), ns).text)
        end = float(node.find("./dataPoints/positions[@axis='{}']/endPosition".format(axis_name), ns).text)
        return np.linspace(start, end, num=numpoints, endpoint=True)

def parse_scan_node(node, ns):
    output = {}
    
    output["scanAxis"] = node.attrib["scanAxis"]

    output['startTimeStamp'] = node.find("./header/startTimeStamp", ns).text

    intensities_node = node.find("./dataPoints/intensities", ns)
    if intensities_node is not None:
        intensities = np.fromstring(intensities_node.text, dtype=float, sep=" ")
        numpoints = len(intensities)
        output["intensities"] = intensities
    else:
        counts_node = node.find("./dataPoints/counts", ns) # later schemas use raw "counts"
        output["counts"] = np.fromstring(counts_node.text, dtype=float, sep=" ")
        numpoints = len(output["counts"])

    for axis in ["2Theta", "Omega", "Phi"]:
        output[axis] = get_positions(node, axis, ns, numpoints)
    
    count_time_node = node.find("./dataPoints/commonCountingTime", ns)
    if (count_time_node is not None):
        count_time = np.ones((numpoints,), dtype=float) * float(count_time_node.text)
    else:
        count_time_node = node.find("./dataPoints/countingTimes", ns)
        count_time = np.fromstring(count_time_node.text, dtype=float, sep=" ")
    output["count_time"] = count_time

    attenuation = 1.0
    attenuation_node = node.find("./dataPoints/commonBeamAttenuationFactor", ns)
    if attenuation_node is not None:
        attenuation = np.ones((numpoints,), dtype=float) * float(attenuation_node.text)
    
    attenuations_node = node.find("./dataPoints/beamAttenuationFactors", ns)
    if attenuations_node is not None:
        attenuation = np.fromstring(attenuations_node.text, dtype=float, sep=" ")
    output["attenuation"] = attenuation

    if not "counts" in output:
        # then we got intensities instead...
        output["counts"] = intensities / attenuation

    return output


"""
To serve with uwsgi:

uwsgi --http 0.0.0.0:8002 --manage-script-name --mount /=server_flask:app

To serve with python:

python server_flask.py 8002
(then visit http://localhost:8002 in your browser)

"""
import os, sys, posixpath
import traceback
import logging
import pkg_resources

from flask import Flask, request, make_response, redirect, send_from_directory
from flask_cors import CORS
from werkzeug.exceptions import HTTPException
import msgpack as msgpack_converter
import json
import mimetypes
mimetypes.add_type("text/css", ".css")
mimetypes.add_type("text/html", ".html")
mimetypes.add_type("application/json", ".json")
mimetypes.add_type("text/javascript", ".js")
mimetypes.add_type("text/javascript", ".mjs")
mimetypes.add_type("image/png", ".png")
mimetypes.add_type("image/svg+xml", ".svg")

def create_app(config=None):
    from reductus.web_gui import api

    RPC_ENDPOINT = '/RPC2'
    STATIC_FOLDER = "webreduce"
    STATIC_PATH = pkg_resources.resource_filename('reductus.web_gui', 'webreduce/')
    PREBUILT_CLIENT = os.path.join("dist", "index.html")
    DEV_CLIENT = "index.html"

    app = Flask(__name__, static_folder=STATIC_PATH, static_url_path='/webreduce')
    CORS(app)
    app.config['SEND_FILE_MAX_AGE_DEFAULT'] = 0

    @app.route('/')
    def root():
        if os.path.exists(os.path.join(STATIC_PATH, PREBUILT_CLIENT)):
            return redirect(posixpath.join(STATIC_FOLDER, PREBUILT_CLIENT))
        else:
            return redirect(posixpath.join(STATIC_FOLDER, DEV_CLIENT))

    @app.route('/robots.txt')
    def static_from_root():
        return send_from_directory(app.static_folder, request.path[1:])

    @app.errorhandler(Exception)
    def handle_error(e):
        code = 500
        if isinstance(e, HTTPException):
            code = e.code
        content = {'exception': repr(e), 'traceback': traceback.format_exc()}
        logging.info(content['traceback'])
        response = make_response(msgpack_converter.packb(content, use_bin_type=True), code)
        response.headers['Content-Type'] = "application/msgpack"
        return response

    def wrap_method(mfunc):
        def wrapper(*args, **kwargs):
            real_kwargs = request.get_json() if request.get_data() else {}
            return_type = request.headers.get("Accept", "application/msgpack")
            if return_type not in ["application/json", "application/msgpack"]:
                code = 406
                content = {'exception': 
                    'no valid Accept return type provided. \
                    (leave unspecified or use one of application/json or application/msgpack)'}
                return_type = "application/json"
                packed = json.dumps(content)
            else:
                content = mfunc(*args, **real_kwargs)
                if return_type == "application/msgpack":
                    packed = msgpack_converter.packb(content, use_bin_type=True)
                else:
                    packed = json.dumps(content)

            response = make_response(packed)
            response.headers['Content-Type'] = return_type
            return response
        return wrapper

    api.initialize(config)

    for method in api.api_methods:
        mfunc = getattr(api, method)
        wrapped = wrap_method(mfunc)
        path = posixpath.join(RPC_ENDPOINT, method)
        shortpath = posixpath.join("/", method)
        app.add_url_rule(path, path, wrapped, methods=["POST"])
        app.add_url_rule(shortpath, shortpath, wrapped, methods=["POST"])

    from reductus.rev import print_revision
    print_revision()

    return app

if __name__ == '__main__':
    logging.basicConfig(level=logging.WARNING)
    port = 8002
    if len(sys.argv) > 1:
        port = int(sys.argv[1])
    app = create_app()
    app.run(port=port)



"""
Extensions to the uncertainties package.

:func:`interp` is a drop-in replacement for numpy.interp which operates on
uncertainty values.

:func:`interp_err` and :func:`solve_err` operate on value and uncertainty in
different variables, but uses the machinery of the uncertainties package to
do the logic.

:func:`format_err` is a helper function for formatting value + uncertainty in
different variables.
"""


from uncertainties import ufloat
from uncertainties.unumpy import \
    uarray, umatrix, nominal_values as uval, std_devs as udev
import numpy as np

def interp_err(x, xp, fp, dfp, left=None, right=None):
    """
    Linear interpolation of x into points (xk,yk +/- dyk).

    xp is assumed to be in ascending order.

    left is the uncertainty value to return for points before the range of xp,
    or None for the initial value, fp[0].

    right is the uncertainty value to return for points after the range of xp,
    or None for the final value, fp[-1].
    """
    is_scalar_x = np.isscalar(x)
    if is_scalar_x:
        fp = ufloat(fp, dfp)
        f = interp([x], xp, fp, left, right)[0]
        return f.n, f.s
    else:
        fp = uarray(fp, dfp)
        f = interp(x, xp, fp, left, right)
        return uval(f), udev(f)

def format_err(x, dx):
    """
    Format number with uncertainty.
    """
    return ufloat(x, dx).format('gS')


def solve_err(A, dA, b, db):
    A = umatrix(A, dA)
    b = umatrix(b, db).T
    x = A.I*b
    return uval(x), udev(x)


def interp(x, xp, fp, left=None, right=None):
    """
    1-D interpolation of *x* into *(xp,fp)*.

    *xp* must be an increasing vector.  *x* can be scalar or vector.

    If *x* is beyond the range of *xp*, returns *left/right*, or the value of
    *fp* at the end points if *left/right* is not defined.

    Implemented in pure python so *fp* can be an extended numeric type such
    as complex or value+uncertainty.
    """
    is_scalar_x = np.isscalar(x)
    if len(xp) == 1:
        f = fp[np.zeros_like(x, dtype='i')]
    else:
        xp = np.asarray(xp)
        if np.any(np.diff(xp) < 0.):
            raise ValueError("interp needs a sorted list")
        if not is_scalar_x:
            x = np.asarray(x)
        idx = np.searchsorted(xp[1:-1], x)
        # Support repeated values in Xp, which will lead to 0/0 errors if the
        # interpolated point is one of the repeated values.
        p = (xp[idx+1]-x)/(xp[idx+1]-xp[idx])
        f = p*fp[idx] + (1-p)*fp[idx+1]

    if is_scalar_x:
        if x < xp[0]:
            return left if left is not None else fp[0]
        elif x > xp[-1]:
            return right if right is not None else fp[-1]
        else:
            return f
    else:
        f[x < xp[0]] = left if left is not None else fp[0]
        f[x > xp[-1]] = right if right is not None else fp[-1]
        return f



from copy import copy

import numpy as np
from numpy import pi, sqrt, polyval, radians, tan, arctan, inf, sin

from reductus.dataflow.lib.wsolve import wpolyfit
from reductus.dataflow.lib.uncertainty import Uncertainty as U, interp
from .util import extend

class FootprintData(object):
    def __init__(self, p, dp):
        self.p = p
        self.dp = dp
    def get_metadata(self):
        #return {"p": self.p.tolist(), "dp": self.dp.tolist()}
        return {
            "slope": self.p[0],
            "intercept": self.p[1],
            # these next are intentionally different from the names of
            # the inputs fields in the footprint module, so that
            # they don't populate those fields when running manually.
            "slope_fit_error_": self.dp[0],
            "intercept_fit_error": self.dp[1]
        }
    def get_plottable(self):
        return {"params": self.get_metadata(), "type": "params"}

def fit_footprint(data, low, high, kind='line'):
    """
    Fit the footprint using data points over a range.

    Range should be a list of pairs of indices, one per data line that is
    to be simultaneously fitted.
    """
    # Join the ranges from the individual data sets
    x, y, dy = [], [], []
    for data_k in data:
        idx = np.ones_like(data_k.Qz, dtype='bool')
        if low is not None:
            idx = idx & (data_k.Qz >= low)
        if high is not None:
            idx = idx & (data_k.Qz <= high)
        x.append(data_k.Qz[idx])
        y.append(data_k.v[idx])
        dy.append(data_k.dv[idx])

    x = np.hstack(x)
    y = np.hstack(y)
    dy = np.hstack(dy)
    if len(x):
        p, dp = _fit_footprint_data(x, y, dy, kind)
        return FootprintData(p, dp)
    else:
        return None


def fit_footprint_shared_range(data, low, high, kind='line'):
    r"""
    Fit the footprint using data points over a range.

    The fit is restricted to *low <= x <= high*.  *kind* can be 'plateau' if
    the footprint is a constant scale factor, 'slope' if the footprint should
    go through the origin, or 'line' if the footprint is a slope that does
    not go through the origin.

    Note: this algorithm is provided for compatibility with the older reflpak
    program.  New datasets need to include support for overlapping points
    with different $\Delta Q$.
    """
    # Join all the datasets
    x, y, dy = [], [], []
    for d in data:
        x.append(d.Qz)
        y.append(d.v)
        dy.append(d.dv)
    x = np.hstack(x)
    y = np.hstack(y)
    dy = np.hstack(dy)


    if low > high:
        low, high = high, low
    idx = (x >= low) & (x <= high)
    x, y, dy = x[idx], y[idx], dy[idx]
    p, dp = _fit_footprint_data(x, y, dy, kind)
    return FootprintData(p, dp)


def apply_fitted_footprint(data, fitted_footprint, range):
    p, dp = fitted_footprint.p, fitted_footprint.dp
    Qmin, Qmax = range
    if Qmax is None:
        Qmax = data.Qz.max()
    if Qmin is None:
        Qmin = data.Qz.min()
    footprint = _generate_footprint_curve(p, dp, data.Qz, Qmin, Qmax)
    _apply_footprint(data, footprint)


def apply_measured_footprint(data, measured_footprint):
    x = measured_footprint.Qz
    y = U(measured_footprint.v, measured_footprint.dv**2)
    footprint = interp(data.Qz, x, y, left=U(1.0, 0.0), right=U(1.0, 0.0))
    _apply_footprint(data, footprint)


def apply_abinitio_footprint(data, Io, width, offset):
    slit1 = (data.slit1.distance, data.slit1.x, data.slit1.y)
    slit2 = (data.slit2.distance, data.slit2.x, data.slit2.y)
    theta = data.sample.angle_x
    if width is None:
        width = data.sample.width
    if offset is None:
        offset = 0.
    if Io is None:
        Io = 1.
    y = Io * _abinitio_footprint(slit1, slit2, theta, width, offset)
    footprint = U(y, 0.0*y)
    _apply_footprint(data, footprint)


def _apply_footprint(data, footprint):
    refl = U(data.v, data.dv**2)
    # Ignore footprint <= 0
    bad_correction = (footprint.x <= 0.)  # type: np.ndarray
    if bad_correction.any():
        footprint = copy(footprint)
        footprint.x = copy(footprint.x)
        footprint.x[bad_correction] = 1.0
        footprint.variance[bad_correction] = 0.0
    corrected_refl = refl/extend(footprint, refl)
    data.v, data.dv = corrected_refl.x, corrected_refl.dx


def _fit_footprint_data(x, y, dy, kind):
    """
    Fit the footprint from the measurement in *x*, *y*, *dy*.

    The fit is restricted to *low <= x <= high*.  *kind* can be 'plateau' if
    the footprint is a constant scale factor, 'slope' if the footprint should
    go through the origin, or 'line' if the footprint is a slope that does
    not go through the origin.
    """
    if len(x) < 2:
        p, dp = np.array([0., 1.]), np.array([0., 0.])
    elif kind == 'plateau':
        poly = wpolyfit(abs(x), y, dy, degree=0, origin=False)
        p, dp = poly.coeff, poly.std
        p, dp = np.hstack((0, p)), np.hstack((0, dp))
    elif kind == 'slope':
        poly = wpolyfit(abs(x), y, dy, degree=1, origin=True)
        p, dp = poly.coeff, poly.std
    elif kind == 'line':
        poly = wpolyfit(abs(x), y, dy, degree=1, origin=False)
        p, dp = poly.coeff, poly.std
    else:
        raise TypeError('unknown footprint type %r'%kind)
    return p, dp


def _generate_footprint_curve(p, dp, x, xmin, xmax):
    """
    Return the footprint correction for the fitted footprint *p*, *dp*.

    The footprint is calculated at the measured points *x*, and applied
    between *xmin* and *xmax*.
    """
    ## order min and max correctly
    xmin, xmax = abs(xmin), abs(xmax)
    if xmin > xmax:
        xmin, xmax = xmax, xmin

    ## linear between Qmin and Qmax
    y = polyval(p, abs(x))
    var_y = polyval(dp**2, x**2)

    ## ignore values below Qmin
    y[abs(x) < xmin] = 1.
    var_y[abs(x) < xmin] = 0.
    ## stretch Qmax to the end of the range
    y[abs(x) > xmax] = polyval(p, xmax)
    var_y[abs(x) > xmax] = polyval(dp**2, xmax**2)

    return U(y, var_y)


def apply(fp, R):
    """
    Scale refl by footprint, ignoring zeros
    """
    x, y, dy = R
    correction = fp.calc()
    correction += (correction == 0)  # avoid divide by zero errors
    corrected_y = y / correction
    corrected_dy = dy / correction
    return corrected_y, corrected_dy


def _abinitio_footprint(slit1, slit2, theta, width, offset=0.):
    """
    Ab-initio footprint calculation from slits, angles and sample size.

    *slit1*, *slit2* are *(d, x, y)* tuples for the distance between the
    slit and the sample, the width of the slit in the beam direction, and
    the width of the slit perpendicular to the beam. The center of slit 2
    must lie on the line between the center of slit 1 and the center of
    rotation of the sample.  All measurements are in mm, though any unit
    will work so long as all measurements are in the same unit.

    *theta* in degrees is the set of angles at which to compute the footprint.

    *width* is the width of the sample. *offset* is a shift in the center of
    rotation of the sample away from the beam center.  These are in the same
    units as the slits (mm).
    """
    # TODO: implement footprint for circular samples
    # TODO: implement vertical footprint if slit y is not inf

    d1, s1x, s1y = slit1
    d2, s2x, s2y = slit2

    # for some reason, distances are stored in nexus files with float32 precision,
    # causing roundoff errors in this calculation
    if hasattr(d1, 'astype'):
        d1 = d1.astype(np.float64)
    if hasattr(d2, 'astype'):
        d2 = d2.astype(np.float64)

    # use radians internally
    theta = radians(theta)

    #use_y = np.all(np.isfinite(s1y) & np.isfinite(s2y))

    # Projection of the sample to the center of rotation
    p_near = (-width / 2 + offset) * sin(theta)
    p_far = (+width / 2 + offset) * sin(theta)

    # beam profile is a trapezoid, 0 where a neutron entering at -s1/2 just
    # misses s2/2, and 1 where a neutron entering at s1/2 just misses s2/2.
    # With tiny front slits, the projection h2 may land on the other side
    # of the beam center.  In this case the overall instensity is lower, but
    # the trapezoid happens to have the same shape, with the flat region
    # extending from -abs(w2) to abs(w2).  Since we only care about intensity
    # relative to the beam for footprint correction, the scale factor cancels.
    w1 = abs(d1/(d1-d2))*(s1x + s2x) / 2 - s1x / 2
    w2 = abs(-abs(d1/(d1-d2)) * (s1x - s2x) / 2 + s1x / 2)
    w_intensity = w1 + w2
    #if use_y:
    #    h1 = abs(d1/(d1-d2))*(s1y+s2y)/2 - s1y/2
    #    h2 = abs(-abs(d1/(d1-d2))*(s1y-s2y)/2 + s1y/2)
    #    h_intensity = h1 + h2

    # Rectangular distribution
    # Three sections of the trapezoid.  Compute the portion of each section
    # that is to the right of p_near and to the right of p_far.  Total is
    # sum the near minus the sum of the far.
    fp = (_trapezoid_sum_right(p_near, w1, w2)
          - _trapezoid_sum_right(p_far, w1, w2))/w_intensity
    return fp

def _trapezoid_sum_right(p, w1, w2):
    """
    Find the integral from p to inf of the trapezoid defined by w1 and w2.
    """
    slope = 1.0/(w1-w2)
    a = np.maximum(p, -w1) + w1
    b = np.maximum(p, -w2) + w1
    A = 0.5*slope*(b**2-a**2)*(b > a)
    a = np.maximum(p, -w2) + w2
    b = np.maximum(p, w2) + w2
    A += (b-a)*(b > a)
    a = w1 - np.maximum(p, w1)
    b = w1 - np.maximum(p, w2)
    A += 0.5*slope*(b**2-a**2)*(b > a)
    return A


def spill(slit, Qz, wavelength, detector_distance, detector_width, thickness,
          A, B, Io, length, offset):
    """
    The primary beam on the detector is the beam reflected from the
    sample. Beam spill is the portion of the beam which is not
    intercepted by the sample but is still incident on the detector.
    This happens at low angles.  Above the sample this is fairly
    simple, being just that portion which is not reflected.  Below
    the sample there is the effect of the the thickness of the
    sample which shades the beam plus the fact that the detector is
    moving out of the path of the beam. At low angles there will also
    be some beam transmitted through the sample, but this is assumed
    to be orders of magnitude smaller than the direct beam so we can
    safely ignore it.  The final effect is the width of the back
    slits which cuts down the transmitted beam.
    """
    if True:
        raise NotImplementedError()

    # It is too difficult to compute beam spill for now.  Leave this
    # pseudo code around in case we decide to implement it later.
    L1 = length/2. + offset
    L2 = length/2. - offset
    wA = slit * A/2.
    wB = slit * B/2.

    # low: low edge of the sample
    # high: high edge of the sample
    # low2: low edge of the sample bottom
    # high2: max(sample,detector)
    # det: low edge of the detector
    # refl: area of intersection
    # spill_lo: area of spill below
    # spill_hi: area of spill above

    # Algorithm for converting Qx-Qz to alpha-beta:
    #   beta = 2 asin(wavelength/(2 pi) sqrt(Qx^2+Qz^2)/2) * 180/pi
    #        = asin(wavelength sqrt(Qx^2+Qz^2) /(4 pi)) / (pi/360)
    #   theta = atan2(Qx,Qz) * 180/pi
    #   alpha = theta + beta/2
    # Since we are in the specular condition, Qx = 0
    #   Qx = 0 => theta => 0 => alpha = beta/2
    #          => alpha = 2 asin(wavelength sqrt(Qz^2)/(4 pi)) / 2
    #          => alpha = asin (wavelength Qz / 4 pi) in radians
    # Length of intersection d = L sin (alpha)
    #          => d = L sin (asin (wavelength Qz / 4 pi))
    #          => d = L wavelength Qz/(4 pi)
    low = -L2*wavelength * Qz / (4 * pi)
    high = L1*wavelength * Qz / (4 * pi)
    area = integrate(wA, wB, low, high)

    # From trig, the bottom of the detector is located at
    #    d sin T - D/2 cos T
    # where d is the detector distance and D is the detector length.
    # Using cos(asin(x)) = sqrt(1-x^2), this is
    #    d wavelength Qz/4pi - D/2 sqrt(1-(wavelength Qz/4pi)^2)
    det = (detector_distance*wavelength*Qz/(4*pi)
           - detector_width/2 * sqrt(1 - (wavelength*Qz/(4*pi))**2))

    # From trig, the projection of thickness in the plane normal to
    # the beam is
    #    thickness/cos(theta) = thickness/sqrt(1-(wavelength Qz/4 pi)^2)
    # since cos(asin(x)) = sqrt(1-x^2).
    low2 = low - thickness / sqrt(1 - (wavelength*Qz/(4*pi))**2)
    high2 = det*(det >= high) + high*(det < high)
    spill_low = integrate(wA, wB, det, low2)
    spill_high = integrate(wA, wB, high2, wB)

    # Total area of intersection is the sum of the areas of the regions
    # Normalize that by the total area of the beam (A+B)/2
    abfoot_y = 2 * Io * (area + spill_low + spill_high) / (wA+wB)
    return abfoot_y



r"""
Background field subtraction
"""
from __future__ import division

import numpy as np
import scipy.optimize as so
from copy import copy

#from reductus.dataflow.lib.uncertainty import Uncertainty as U, interp
#from .refldata import ReflData

# BackgroundFieldData class returns results of background field fitting

class BackgroundFieldData(object):
    def __init__(self, p, dp2, s, ds2, dsp, chisq):
        self.p = p
        self.dp2 = dp2
        self.s = s
        self.ds2 = ds2
        self.dsp = dsp
        self.chisq = chisq
    def get_metadata(self):
        return {
            "epsD": self.p,
            "epsD_var": self.dp2,
            "scale": self.s,
            "scale_var": self.ds2,
            "scale_epsD_covar": self.dsp,
            "fit_chi-squared": self.chisq
        }
    def get_plottable(self):
        return {"params": self.get_metadata(), "type": "params"}


def background_reservoir(ai, epsD, af, epssi, Fd, scale):
    """
    Calculates background from a reservoir surrounded by Si
    """
    with np.errstate(divide='ignore'):
        A = np.exp(-epsD * (1. / np.sin(ai) + 1. / np.sin(af)))
        res = (1 - A) / (1. + np.sin(ai) / np.sin(af))
        #si = epssi * (Fd * np.sin(af) / np.sin(ai + af)) * (0.5 + 0.5 * A)
        si = epssi * Fd / np.cos(ai) * (0.5 + 0.5 * A)

    # first term is the background value; the second contains intermediates used
    # in the Jacobian function and error calculation
    return scale*(res+si), (A, res, si)

def background_reservoir_jac(ai, epsD, af, epssi, Fd, scale):
    """
    Calculates Jacobian function for background_reservoir with
    parameters (scale, epsD)
    """
    A, res, si = background_reservoir(ai, epsD, af, epssi, Fd, scale)[1]

    with np.errstate(divide='ignore'):
        dA_depsD = -(1. / np.sin(ai) + 1. / np.sin(af)) * A
        dres_depsD = -dA_depsD / (1. + np.sin(ai) / np.sin(af))
        #dsi_depsD = epssi * (Fd * np.sin(af) / np.sin(ai + af)) * (0.5 * dA_depsD)
        dsi_depsD = epssi * Fd / np.cos(ai) * (0.5 * dA_depsD)

    J = np.array([res + si, scale*(dres_depsD + dsi_depsD)])

    return J

def background_reservoir_error(ai, epsD, af, epssi, Fd, scale, pcov):
    """
    Uses covariance matrix of fit result to propagate uncertainty
    """
    J = background_reservoir_jac(ai, epsD, af, epssi, Fd, scale)

    return np.sqrt(np.array([j.T.dot(pcov.dot(j)) for j in J.T]))

def fractional_solid_angle(s4, LS4, L4D, HD):
    """
    Calculates fractional solid angle subtended by a detector
    of height HD located a distance LS4 from the sample, fronted
    by a collimating slit a distance L4D away with width s4. Only
    works if the result is << 1.
    """

    dvert = np.arctan( HD / (LS4 + L4D) )
    dhorz = np.arctan( s4 / LS4 )
    return dvert * dhorz / (4 * np.pi)

def detector_footprint(af, s3, s4, LS3, L34):
    """
    Calculates effective footprint on sample viewed by the detector,
    based on post-sample slit values and the angle between the sample
    and the detector.
    """

    return s3 * LS3 / (L34 * np.sin(af)) * (s4 / s3 + 1. + L34 / LS3)

def fit_background_field(back, epsD0, epssi, fit_scale, scale_value=1.0, LS3=380, LS4=1269, LSD=1675, HD=150, Qcutoff=0.05, maxF=75):
    """
    Performs the fit to the background field.
    """

    ##type: (ReflData) -> (np.ndarray)
    #TODO: Better error handling if slit 4 doesn't exist (currently hardcoded detector size)
    #TODO: Include residuals for inspection (can currently pass both through a subtract module)

    # Initialize variables
    ai = list()
    af = list()
    s3 = list()
    s4 = list()
    v = list()
    dv = list()
    Qz_target = list()

    # Check for slit distances, otherwise use values from data file
    LS3 = back[0].slit3.distance if back[0].slit3.distance is not None else LS3
    LS4 = back[0].slit4.distance if back[0].slit4.distance is not None else LS4
    L4D = LSD - LS4

    # Extract relevant quantities from inputs
    for b in back:
        ai.append(b.sample.angle_x)
        Qz_target.append(b.Qz_target if not np.all(np.isnan(b.Qz_target))
                         else 4*np.pi/b.detector.wavelength*np.sin(np.radians(b.sample.angle_x)))
        af.append(b.detector.angle_x - b.sample.angle_x)
        s3.append(b.slit3.x)
        s4.append(b.slit4.x if not np.all(np.isinf(b.slit4.x)) else 25. * np.ones(b.v.shape))
        v.append(b.v)
        dv.append(b.dv)

    # Concatenate and filter results based on the Qcutoff parameter
    Qz_target = np.concatenate(Qz_target)
    ai = np.radians(np.concatenate(ai))
    af = np.radians(np.concatenate(af))
    crit = (Qz_target > Qcutoff) & (af > 0.)
    ai = ai[crit]
    af = af[crit]
    s3 = np.concatenate(s3)[crit]
    s4 = np.concatenate(s4)[crit]
    v = np.concatenate(v)[crit]
    dv = np.concatenate(dv)[crit]

    # Calculate the detector footprints and use maximum value relative to sample size maxF
    Fd = detector_footprint(af, s3, s4, LS3, LS4 - LS3)
    Fd = np.min((Fd, maxF*np.ones(Fd.shape)), axis=0)

    # Calculate the detector solid angle
    SA = fractional_solid_angle(s4, LS4, L4D, HD)

    # Define fit function and Jacobian
    def minfunc(pars):
        scale, epsD = pars
        res = background_reservoir(ai, epsD, af, epssi, Fd, scale)
        return (res[0] * SA - v)/dv

    def jacfunc(pars):
        scale, epsD = pars
        J = background_reservoir_jac(ai, epsD, af, epssi, Fd, scale)
        return (J*np.tile(SA/dv, (len(pars), 1))).T

    # Condition the fit if the scale factor is not included in the fit, then perform the fit
    if not fit_scale:
        pout, pcov = so.leastsq(
            lambda epsD: minfunc([scale_value, epsD]),
            epsD0,
            Dfun=lambda epsD: jacfunc([1.0, epsD])[:,1],
            full_output=True, ftol=1e-15, xtol=1e-15)[:2]
        lenpout = len(pout)
        pout = np.array([scale_value, pout[0]])
        pcov = np.array([[0.0, 0.0], [0.0, pcov[0][0]]])
    else:
        pout, pcov = so.leastsq(
            minfunc,
            np.array([scale_value, epsD0]),
            Dfun=jacfunc,
            full_output=True, ftol=1e-15, xtol=1e-15)[:2]
        lenpout = len(pout)

    # Calculate fitting error (not used) and chi-squared (only reported)
    perr, chisq = np.sqrt(np.diag(pcov)), np.sum(minfunc(pout)**2) / (len(v) - lenpout)

    # for testing purposes only
    if 0:
        import matplotlib.pyplot as plt
        plt.errorbar(np.arange(len(v)), v, dv, fmt='o')
        res = background_reservoir(ai, pout[1], af, epssi, Fd, pout[0])
        plt.plot(np.arange(len(v)), res[0] * SA)
        plt.show()

    # Create fit data structure
    bff = BackgroundFieldData(pout[1], pcov[1,1], pout[0], pcov[0,0], pcov[0,1], chisq)
    bff.epssi = epssi
    bff.LS3 = LS3
    bff.LS4 = LS4
    bff.L4D = L4D
    bff.HD = HD
    bff.maxF = maxF

    # Create calculated background data for comparison to input background data
    back2 = [copy(b) for b in back]

    for i, b in enumerate(back2):
        ai2 = np.radians(b.sample.angle_x)
        af2 = np.radians(b.detector.angle_x - b.sample.angle_x)
        s32 = b.slit3.x
        s42 = b.slit4.x if not np.all(np.isinf(b.slit4.x)) else 25. * np.ones(b.v.shape)

        # Uncomment following lines if output data should only include the data above Qcutoff
        # Currently all data are calculated

        #Qz_target = b.Qz_target if not np.all(np.isnan(b.Qz_target)) else 4 * np.pi / b.detector.wavelength * np.sin(
        #        np.radians(b.sample.angle_x))
        #crit = (Qz_target > Qcutoff) & (af > 0.)
        #ai2 = ai2[crit]
        #af2 = af2[crit]
        #s32 = s32[crit]
        #s42 = s42[crit]

        Fd2 = detector_footprint(af2, s32, s42, LS3, LS4 - LS3)
        Fd2 = np.min((Fd2, maxF * np.ones(Fd2.shape)), axis=0)
        SA2 = fractional_solid_angle(s42, LS4, L4D, HD)

        bestfit = background_reservoir(ai2, pout[1], af2, epssi, Fd2, pout[0])[0]
        bestfiterr = background_reservoir_error(ai2, pout[1], af2, epssi, Fd2, pout[0], pcov)

        back2[i].v = bestfit * SA2
        back2[i].dv = bestfiterr * SA2

    return bff, back2

def apply_background_field_subtraction(data, epsD, epssi, LS3, LS4, L4D, HD, maxF, scale, pcov):
    """
    Applies background field subtraction to specular data
    """
    for (i,d) in enumerate(data):
        I, varI = subtract_background_field(d, epsD, epssi, LS3, LS4, L4D, HD, maxF, scale, pcov)
        data[i].v, data[i].dv = I, np.sqrt(varI)

def subtract_background_field(d, epsD, epssi, LS3, LS4, L4D, HD, maxF, scale, pcov):
    """
    Subtracts background field from a single specular data set with error propagation
    """
    # TODO: Better error handling if slit 4 doesn't exist (currently hardcoded detector size)
    ai = np.radians(d.sample.angle_x)
    af = np.radians(d.detector.angle_x - d.sample.angle_x)
    s4 = d.slit4.x if not np.all(np.isinf(d.slit4.x)) else 25. * np.ones(d.v.shape)
    Fd = detector_footprint(af, d.slit3.x, s4, LS3, LS4 - LS3)
    Fd = np.min((Fd, maxF * np.ones(Fd.shape)), axis=0)
    SA = fractional_solid_angle(s4, LS4, L4D, HD)

    bkg = background_reservoir(ai, epsD, af, epssi, Fd, scale)[0]*SA
    bkg_err = background_reservoir_error(ai, epsD, af, epssi, Fd, scale, pcov)*SA

    return d.v - bkg, d.dv**2 + bkg_err**2



from __future__ import print_function

import sys
import posixpath
import zipfile
import json

import numpy

def bytes_to_str(s):
    # type: (AnyStr) -> str
    return s.decode('utf-8') if isinstance(s, bytes) else s  # type: ignore

builtin_open = open

__version__ = "0.0.1"


class Node(object):
    _attrs_filename = ".attrs"

    def __init__(self, parent_node=None, path="/", nxclass=None, attrs=None):
        self.root = self if parent_node is None else parent_node.root
        self.readonly = self.root.readonly
        if path.startswith("/"):
            # absolute path
            self.path = path
        else:
            # relative
            self.path = posixpath.join(parent_node.path, path)

    def makeAttrs(self):
        attr_data = self.root.open(posixpath.join(self.path, self._attrs_filename), "r").read()
        return json.loads(bytes_to_str(attr_data))

    @property
    def parent(self):
        return self.root[posixpath.dirname(self.name)]

    @property
    def groups(self):
        return dict([(gn, Group(self, gn)) for gn in self.groupnames])

    @property
    def groupnames(self):
        return  [x for x in self.root.listdir(self.path) if self.root.isdir(posixpath.join(self.path, x))]

    @property
    def name(self):
        return self.path

    def keys(self):
        return [x for x in self.root.listdir(self.path)
                if not x.endswith('.attrs') and not x.endswith('.link')]

    def values(self):
        keys = self.keys()
        return [self[k] for k in keys]

    def items(self):
        keys = self.keys()
        return [(k, self[k]) for k in keys]

    def __contains__(self, key):
        return self.root.exists(posixpath.join(self.path, key))

    def __getitem__(self, path):
        """ get an item based only on its path.
        Can assume that next-to-last segment is a group (dataset is lowest level)
        """
        if path.startswith("/"):
            # absolute path
            full_path = path
        else:
            # relative
            full_path = posixpath.join(self.path, path)

        #os_path = posixpath.join(self.os_path, full_path.lstrip("/"))
        if self.root.exists(full_path):
            if self.root.isdir(full_path):
                # it's a group
                return Group(self, full_path)
            elif self.root.exists(full_path + ".link"):
                # it's a link
                return makeSoftLink(self, full_path)
            else:
                # it's a field
                return FieldFile(self, full_path)
        else:
            # the item doesn't exist
            raise KeyError(path)

    def get(self, path, default_value):
        try:
            value = self.__getitem__(path)
            return value
        except KeyError:
            return default_value

    def add_field(self, path, **kw):
        FieldFile(self, path, **kw)

    def add_group(self, path, nxclass, attrs={}):
        Group(self, path, nxclass, attrs)


class File(Node):
    def __init__(self, filename, file_obj=None):
        self.readonly = True
        Node.__init__(self, parent_node=None, path="/")
        if file_obj is None:
            file_obj = builtin_open(filename, mode='rb')
        self.zipfile = zipfile.ZipFile(file_obj)
        self.attrs = self.makeAttrs()
        self.filename = filename
        self.mode = "r"

    def flush(self):
        # might make this do writezip someday.
        pass

    def isdir(self, path):
        """ abstraction for looking up paths:
        should work for unpacked directories and packed zip archives """
        path = path.lstrip("/")
        if path == "":
            return True # root path
        else:
            filenames = self.root.zipfile.namelist()
            return (path.rstrip("/") + "/") in filenames

    def listdir(self, path):
        """ abstraction for looking up paths:
        should work for unpacked directories and packed zip archives """
        path = path.strip("/")
        return [posixpath.basename(fn.rstrip("/"))
                for fn in self.zipfile.namelist()
                if posixpath.dirname(fn.rstrip("/")) == path]

    def exists(self, path):
        path = path.strip("/")
        filenames = self.root.zipfile.namelist()
        return path in filenames or self.isdir(path)

    def read(self, path):
        return self.open(path, "r").read()

    def getsize(self, path):
        path = path.lstrip("/")
        return self.zipfile.getinfo(path).file_size

    def open(self, path, mode):
        path = path.lstrip("/")
        return self.zipfile.open(path, "r")

    def __repr__(self):
        return "<HDZIP file \"%s\" (mode %s)>" % (self.filename, self.mode)

    def close(self):
        # there seems to be only one read-only mode
        self.zipfile.close()


class Group(Node):
    def __init__(self, node, path, nxclass="NXCollection", attrs={}):
        Node.__init__(self, parent_node=node, path=path)
        if path.startswith("/"):
            # absolute path
            self.path = path
        else:
            # relative
            self.path = posixpath.join(node.path, path)

        self.attrs = self.makeAttrs()

    def __repr__(self):
        return "<HDZIP group \"" + self.path + "\">"


class FieldFile(object):
    _formats = {
        'S': '%s',
        'f': '%.8g',
        'i': '%d',
        'u': '%d',
        'b': '%d'}

    _attrs_suffix = ".attrs"

    def __init__(self, node, path, **kw):
        self.root = node.root
        if not path.startswith("/"):
            # relative path:
            path = posixpath.join(node.path, path)
        self.path = path
        attr_data = self.root.open(self.path + self._attrs_suffix, "r").read()
        self.attrs = json.loads(bytes_to_str(attr_data))
        self._value = None


    def __repr__(self):
        return ("<HDZIP field \"%s\" %s \"%s\">"
                % (self.name, str(self.attrs['shape']), self.attrs['dtype']))

    def __getitem__(self, slice_def):
        return self.value.__getitem__(slice_def)

    def __setitem__(self, slice_def, newvalue):
        raise Exception("read only")

    # promote a few attrs items to python object attributes:
    @property
    def shape(self):
        return self.attrs.get('shape', None)

    @property
    def dtype(self):
        return self.attrs.get('dtype', None)

    @property
    def name(self):
        return self.path


    @property
    def parent(self):
        return self.root[posixpath.dirname(self.name)]

    @property
    def value(self):
        if self._value is None:
            attrs = self.attrs
            target = self.path
            try:
                infile = self.root.open(target, 'rb')
                dtype_str = str(attrs['format'])
                # CRUFT: <l4, <d8 are not sensible dtypes
                if dtype_str == '<l4': dtype_str = '<i4'
                if dtype_str == '<l8': dtype_str = '<i8'
                if dtype_str == '<d8': dtype_str = '<f8'
                dtype_str = dtype_str.replace('S', 'U')
                dtype = numpy.dtype(dtype_str)
                if attrs.get('binary', False) == True:
                    s = infile.read()
                    d = numpy.frombuffer(s, dtype=dtype).copy()
                elif self.root.getsize(target) == 0:
                    d = numpy.array([], dtype=dtype)
                elif self.root.getsize(target) == 1:
                    # empty entry: only contains \n
                    # this is only possible with empty string being written.
                    d = numpy.array([''], dtype=dtype)
                elif dtype.kind == 'S':
                    data = [[v for v in line[:-1].split(b'\t')]
                            for line in infile]
                    d = numpy.squeeze(numpy.array(data))
                    d = _unescape_str(d)
                elif dtype.kind == 'U':
                    data = [[v.decode('utf-8') for v in line[:-1].split(b'\t')]
                            for line in infile]
                    d = numpy.squeeze(numpy.array(data))
                    d = _unescape_str(d)
                else:
                    d = numpy.loadtxt(infile, dtype=dtype, delimiter='\t')
            finally:
                infile.close()
            if 'shape' in attrs:
                try:
                    d = d.reshape(attrs['shape'])
                except Exception:
                    # literally do nothing.  Should be logging this.
                    pass
            self._value = d
        return self._value

def _unescape_str(data):
    if data.size:
        # Hide the \\ in \1 so that it doesn't get processed twice.  At the
        # end, convert it back to \\.  This allows us to support \\t without
        # it turning into a tab character or a backslash tab sequence.
        # Don't convert empty arrays since the change type
        data = numpy.char.replace(data, '\\\\', '\1')
        data = numpy.char.replace(data, '\\t', '\t')
        data = numpy.char.replace(data, '\\r', '\r')
        data = numpy.char.replace(data, '\\n', '\n')
        data = numpy.char.replace(data, '\1', '\\')

    return data

def makeSoftLink(parent, path):
    orig_attrs_path = path.lstrip("/") + ".link"
    attr_data = parent.root.open(orig_attrs_path, 'r').read()
    linkinfo = json.loads(bytes_to_str(attr_data))
    target = linkinfo['target']
    target_obj = parent.root[target]
    target_obj.orig_path = path
    return target_obj

#compatibility with h5nexus:
group = Group
field = FieldFile
open = File



import numpy as np
import functools

from reductus.dataflow.lib import err1d
from .util import extend


def apply_rescale(data, scale, dscale):
    I, varI = err1d.mul(data.v, data.dv**2, scale, dscale**2)
    data.v, data.dv = I, np.sqrt(varI)

# from https://stackoverflow.com/questions/31174295/getattr-and-setattr-on-nested-subobjects-chained-properties/54547158#54547158
def rgetattr(obj, path, *default):
    try:
        return functools.reduce(getattr, path.split('.'), obj)
    except AttributeError:
        if default:
            return default[0]
        raise

def apply_intensity_norm(data, base, align_by=None):
    assert data.normbase == base.normbase, "can't mix time and monitor normalized data"

    if align_by is None:
        # use the one suggested by the data structure, or error out if missing.
        align_by = getattr(data, 'align_intensity')
    # use "slit1.x" for 'align_by' with Candor data
    data_x, base_x = rgetattr(data, align_by), rgetattr(base, align_by)
    
    S, varS = err1d.interp(data_x, base_x, base.v, base.dv**2)
    # Candor may have only one detector bank active, and so the other may
    # have zeros in it.  Ignore those channels.
    I, varI = err1d.div(data.v, data.dv**2, S + (S==0), varS)
    data.v, data.dv = I, np.sqrt(varI)

def calculate_number(data, base, time_uncertainty=1e-6):
    """ returns the measured base flux * count time for each point """
    assert base.normbase == 'time', "can't calculate time-integrated flux from monitor-normalized base"
    if data.angular_resolution.ndim == 1:
        data_x, base_x = data.angular_resolution, base.angular_resolution
    else:
        data_x, base_x = data.slit1.x, base.slit1.x
    S, varS = err1d.interp(data_x, base_x, base.v, base.dv**2)
    F, varF = err1d.mul(data.monitor.count_time, time_uncertainty**2, S, varS)
    return F, varF


def estimate_attenuation(datasets):
    raise NotImplementedError()
    index = np.sort([d.angular_resolution[0] for d in datasets])


NORMALIZE_OPTIONS = 'auto|monitor|time|roi|power|none'
def apply_norm(data, base='auto'):
    if base == 'auto':
        # We are ignoring counter.countAgainst since monitor is almost
        # always the best choice for normalization.  Even when counting
        # against time monitor normalization protects against flucuations
        # in beam intensity.  Only reason not to use it is when the monitor
        # is bad or missing.
        if data.monitor.counts is not None and (data.monitor.counts > 0).any():
            base = 'monitor'
        elif data.monitor.count_time is not None and (data.monitor.count_time > 0).any():
            base = 'time'
        else:
            base = 'none'

    C = data.detector.counts
    varC = data.detector.counts_variance
    if base == 'monitor':
        #assert (data.monitor.counts > 0).all(), "monitor counts are zero; can't normalize by monitor"
        M = data.monitor.counts
        M[M == 0] = 1  # protect against zero counts in monitor
        varM = data.monitor.counts_variance
        varM[varM == 0] = 1  # variance on zero counts is +/- 1
        units = 'monitor'
    elif base == 'time':
        #assert (data.monitor.count_time > 0).all(), "count time is zero; can't normalize by time"
        M = data.monitor.count_time
        M[M == 0] = data.monitor.time_step/2.  # protect against zero count time
        # Uniform distribution has variance of (interval_width**2)/12
        varM = (data.monitor.time_step**2)/12.
        units = 'second'
    elif base == 'roi':
        M = data.monitor.roi_counts
        M[M == 0] = 1  # protect against zero counts in monitor
        varM = data.monitor.roi_variance
        varM[varM == 0] = 1  # variance on zero counts is +/- 1
        units = 'roi count'
    elif base == 'power':
        if data.monitor.source_power is None:
            raise ValueError("source power is unknown; can't normalize by power")
        # Power is in megawatts, not megawatt-hours, so scale by time.
        time = data.monitor.count_time/3600.
        M = data.monitor.source_power * time
        varM = data.monitor.source_power_variance * time
        units = data.monitor.source_power_units + "-hour"
    elif base == 'none':
        M = 1
        varM = 0
        units = ''
    else:
        raise ValueError("Expected %r in %s" % (base, NORMALIZE_OPTIONS))
    # Broadcast for nD detector arrays
    if C.ndim > 1:
        M, varM = extend(M, C), extend(varM, C)
    #print "norm",C,varC,M,varM
    value, variance = err1d.div(C, varC+(varC == 0), M, varM)
    data.v = value
    data.dv = np.sqrt(variance)
    data.vunits = 'counts per '+units if units else 'counts'
    data.vlabel = 'Intensity'
    data.normbase = base

    # TODO: set data scale to a nice number
    #data.scale = nice_number(M[0])




import os, sys, types
from copy import deepcopy
from posixpath import basename, join
import time
from io import BytesIO
from typing import List

import pytz
from numpy import (cos, pi, cumsum, arange, ndarray, ones, zeros, array,
                   newaxis, linspace, empty, resize, sin, allclose, zeros_like,
                   linalg, dot, arctan2, float64, histogram2d, sum, nansum,
                   sqrt, loadtxt, searchsorted, nan, logical_not, fliplr,
                   flipud, indices, polyfit, radians, argsort)

from numpy.ma import MaskedArray

from reductus.dataflow.lib.h5_open import h5_open_zip

from reductus.dataflow.core import Template
from reductus.dataflow.calc import process_template
from reductus.dataflow.fetch import url_get
from reductus.dataflow.lib import rebin as reb
from reductus.dataflow.lib.iso8601 import seconds_since_epoch

from reductus.dataflow.automod import cache, nocache, module

from .FilterableMetaArray import FilterableMetaArray as MetaArray

DEBUG = False

def get_index(t, x):
        if (x == "" or x == None):
            return None
        if float(x) > t.max():
            return None
        if float(x) < t.min():
            return None
        tord = argsort(t)
        return tord[searchsorted(t, float(x), sorter=tord)]

@module
def fitPSDCalibration(calibration, minimum_peak_intensity=500.0):
    """
    Takes a loaded psd calibration file, which
    is direct beam measurements at a variety of detector angles with tight slits
    X = np.arange(data.size)
    x = np.sum(X*data)/np.sum(data)
    width = np.sqrt(np.abs(np.sum((X-x)**2*data)/np.sum(data)))

    **Inputs**

    calibration (ospec2d) : data in

    minimum_peak_intensity (float): don't fit peaks with integrated intensity smaller than this

    **Returns**

    output (params) : fit results

    2016-04-01 Brian Maranville
    """
    from .dataflow import Parameters
    counts = calibration['Measurements':'counts']
    twotheta = calibration.extrainfo['det_angle']
    cpixels = []
    tts = []
    widths = []
    for i in range(counts.shape[1]):
        data = counts[:,i].view(ndarray)
        tt  = twotheta[i]
        if sum(data) < minimum_peak_intensity:
            continue
        X = arange(data.size) + 0.5
        x = sum(X*data)/sum(data)
        width = sqrt(abs(sum((X-x)**2*data)/sum(data)))
        widths.append(width)
        cpixels.append(x)
        tts.append(tt)
    # now fitting to polynomial:
    fit = polyfit(tts, cpixels, 1)
    # returns [pixels_per_degree, qzero_pixel]
    return Parameters({"pixels_per_degree": -fit[0], "qzero_pixel": fit[1]})

@module
def coordinateOffset(data, axis=None, offset=0):
    """
    Apply an offset to one or both of the coordinate axes
    To apply an offset to an axis, add it to the dict of offsets.
    e.g. if data is theta, twotheta, then
    to apply 0.01 offset to twotheta only make offsets = {'twotheta': 0.01}

    **Inputs**

    data (ospec2d) : data in

    axis (str): axis to apply the offset to

    offset (float): amount of offset

    **Returns**

    output (ospec2d) : data with offsets applied

    2016-04-01 Brian Maranville
    """
    if axis is None: return data
    new_info = data.infoCopy()
    axisnum = data._getAxis(axis)
    new_info[axisnum]['values'] += offset
    new_data = MetaArray(data.view(ndarray).copy(), info=new_info)
    return new_data

@module
def normalizeToMonitor(data):
    """
    divide all the counts columns by monitor and output as normcounts, with stat. error

    **Inputs**

    data (ospec2d) : data in

    **Returns**

    output (ospec2d) : data with normalization applied

    2016-04-01 Brian Maranville
    """
    cols = [col['name'] for col in data._info[-2]['cols']]
    passthrough_cols = [col for col in cols if (not col.startswith('counts'))] # and not col.startswith('monitor'))]
    counts_cols = [col for col in cols if col.startswith('counts')]
    monitor_cols = [col for col in cols if col.startswith('monitor')]
    info = data.infoCopy()
    info[-2]['cols'] = []
    output_array = zeros( data.shape[:-1] + (len(counts_cols) + len(passthrough_cols),), dtype=float) * nan
    expressions = []
    for i, col in enumerate(passthrough_cols):
        info[-2]['cols'].append({"name":col})
        output_array[..., i] = data["Measurements":col]

    for i, col in enumerate(counts_cols):
        j = i + len(passthrough_cols)
        col_suffix = col[len('counts'):]
        monitor_id = 'monitor'
        if ('monitor'+col_suffix) in monitor_cols:
            monitor_id += col_suffix
        info[-2]['cols'].append({"name":"counts_norm%s" % (col_suffix,)})
        mask = data["Measurements":monitor_id].nonzero()
        #print(mask)
        output_array[..., j][mask] = data["Measurements":col][mask] / data["Measurements":monitor_id][mask]
        #expression = "data1_counts%s / data1_%s" % (col_suffix, monitor_id)
        #error_expression = "sqrt(data1_counts%s) / data1_%s" % (col_suffix, monitor_id)
        #expressions.append({"name": "counts_norm%s" % (col_suffix,), "expression":expression})
        #expressions.append({"name": "error_counts_norm%s" % (col_suffix,), "expression":error_expression})
    #result = Algebra().apply(data, None, expressions, passthrough_cols)
    result = MetaArray(output_array, info=info)
    return result

@module
def cropData(data, cropbox=[None,None,None,None]):
    """
    crop 2d data along both axes and return smaller dataset

    **Inputs**

    data (ospec2d) : data in

    cropbox (range?:xy): region over which to mask (in data coordinates)

    **Returns**

    output (ospec2d) : data with normalization applied

    2018-04-10 Brian Maranville
    """
    if cropbox is None:
        cropbox = [None, None, None, None]
    xmin, xmax, ymin, ymax = cropbox
    new_info = data.infoCopy()
    x_axis = new_info[0]
    y_axis = new_info[1]
    col_info = new_info[2]
    extra_info = new_info[3]

    x_array = data._info[0]['values']
    y_array = data._info[1]['values']

    xslice = slice(get_index(x_array, xmin), get_index(x_array, xmax))
    yslice = slice(get_index(y_array, ymin), get_index(y_array, ymax))
    dataslice = (xslice, yslice)
    output_array = data.view(ndarray)[dataslice]
    new_info[0]['values'] = x_array[xslice]
    new_info[1]['values'] = y_array[yslice]
    result = MetaArray(output_array, info=new_info)
    return result

@module
def maskData(data, maskbox=[None,None,None,None], invert=False):
    """
    Set all data, normalization to zero within mask

    **Inputs**

    data (ospec2d) : data in
    
    maskbox (range?:xy): region over which to mask (in data coordinates)
    
    invert (bool): if True, mask everything not in the box

    **Returns**

    masked (ospec2d) : data out, with mask applied

    2018-04-01 Brian Maranville
    """
    
    if maskbox is None:
        maskbox = [None, None, None, None]
    xmin, xmax, ymin, ymax = maskbox
    new_info = data.infoCopy()
    x_axis = new_info[0]
    y_axis = new_info[1]
    col_info = new_info[2]
    extra_info = new_info[3]

    x_array = data._info[0]['values']
    y_array = data._info[1]['values']

    xslice = slice(get_index(x_array, xmin), get_index(x_array, xmax))
    yslice = slice(get_index(y_array, ymin), get_index(y_array, ymax))
    dataslice = (xslice, yslice)
    if invert:
        new_data = MetaArray(zeros_like(data.view(ndarray)), info=data.infoCopy())
        new_data.view(ndarray)[dataslice] = data.view(ndarray).copy()[dataslice]
    else:
        new_data = MetaArray(data.view(ndarray).copy(), info=data.infoCopy())
        new_data.view(ndarray)[dataslice] = 0

    return new_data

@module
def sliceData(data, slicebox=[None,None,None,None]):
    """
    Sum 2d data along both axes and return 1d datasets

    **Inputs**

    data (ospec2d) : data in
    
    slicebox (range?:xy): region over which to integrate (in data coordinates)

    **Returns**

    xout (ospecnd) : xslice

    yout (ospecnd) : yslice

    2016-04-01 Brian Maranville
    """
    
    if slicebox is None:
        slicebox = [None, None, None, None]
    xmin, xmax, ymin, ymax = slicebox
    new_info = data.infoCopy()
    x_axis = new_info[0]
    y_axis = new_info[1]
    col_info = new_info[2]
    extra_info = new_info[3]

    x_array = data._info[0]['values']
    y_array = data._info[1]['values']

    xslice = slice(get_index(x_array, xmin), get_index(x_array, xmax))
    yslice = slice(get_index(y_array, ymin), get_index(y_array, ymax))
    dataslice = (xslice, yslice)
    # print xmin, xmax, ymin, ymax
    x_out = nansum(data.view(ndarray)[dataslice], axis=1)
    y_out = nansum(data.view(ndarray)[dataslice], axis=0)
    x_axis['values'] = x_axis['values'][xslice]
    y_axis['values'] = y_axis['values'][yslice]

    x_data_obj = MetaArray( x_out, info=[x_axis, col_info, extra_info] )
    y_data_obj = MetaArray( y_out, info=[y_axis, col_info, extra_info] )

    return x_data_obj, y_data_obj

@module
def pixelsToTwotheta(data, params, pixels_per_degree=50.0, qzero_pixel=149.0, instr_resolution=1e-6, ax_name='xpixel'):
    """ input array has axes theta and pixels:
    output array has axes theta and twotheta.

    Pixel-to-angle conversion is arithmetic (pixels-per-degree=constant)
    output is rebinned to fit in a rectangular array if detector angle
    is not fixed.

    **Inputs**

    data (ospec2d) : data in

    params (params): parameters override the field values

    pixels_per_degree {Pixels per degree} (float): slope of equation relating pixel to angle

    qzero_pixel {Q-zero pixel} (float): pixel value for Q=0

    instr_resolution {Resolution} (float): steps in angle smaller than this will be ignored/combined

    ax_name {Axis name} (str): name of the axis containing pixel data

    **Returns**

    output (ospec2d) : data with pixel axis converted to angle

    2016-04-01 Brian Maranville
    """

    if params is not None:
        if 'pixels_per_degree' in params: pixels_per_degree = params['pixels_per_degree']
        if 'qzero_pixel' in params: qzero_pixel = params['qzero_pixel']
    #kw = locals().keys()
    #print(kw, params)
    #for name in kw:
    #    if name in params:
    #        exec "print('%s', %s, params['%s']" % (name, name,name)) in locals()
    #        exec ("%s = params['%s']" % (name, name)) in locals()
    #        exec "print(name)" in locals()

    pixels_per_degree = float(pixels_per_degree) # coerce, in case it was an integer
    qzero_pixel = float(qzero_pixel)
    instr_resolution = float(instr_resolution)

    print(pixels_per_degree, qzero_pixel)

    new_info = data.infoCopy()
    det_angle = new_info[-1].get('det_angle', None)
    det_angle = array(det_angle)
    # det_angle should be a vector of the same length as the other axis (usually theta)
    # or else just a float, in which case the detector is not moving!
    ndim = len(new_info) - 2 # last two entries in info are for metadata
    pixel_axis = next((i for i in range(len(new_info)-2) if new_info[i]['name'] == ax_name), None)
    if pixel_axis < 0:
        raise ValueError("error: no %s axis in this dataset" % (ax_name,))

    if hasattr(det_angle, 'max'):
        det_angle_max = det_angle.max()
        det_angle_min = det_angle.min()
    else: # we have a number
        det_angle_max = det_angle_min = det_angle

    if ((det_angle_max - det_angle_min) < instr_resolution) or ndim == 1 or ax_name != 'xpixel':
        #then the detector is fixed and we just change the values in 'xpixel' axis vector to twotheta
        # or the axis to be converted is y, which doesn't move in angle.
        print("doing the simple switch of axis values...")

        #data_slices = [slice(None, None, 1), slice(None, None, 1)]
        #data_slices[pixel_axis] = slice(None, None, -1)

        if ax_name == 'xpixel':
            twotheta_motor = det_angle_min
            new_info[pixel_axis]['name'] = 'twotheta'
        else:
            twotheta_motor = 0.0 # we don't have a y-motor!
            new_info[pixel_axis]['name'] = 'twotheta_y'

        pixels = new_info[pixel_axis]['values']
        twoth = (pixels - qzero_pixel) / pixels_per_degree + twotheta_motor
        #new_info[pixel_axis]['values'] = twoth[::-1] # reverse: twotheta increases as pixels decrease
        new_info[pixel_axis]['values'] = twoth
        new_info[pixel_axis]['units'] = 'degrees'
        #new_array = (data.view(ndarray).copy())[data_slices]
        new_array = (data.view(ndarray).copy())
        new_data = MetaArray(new_array, info=new_info)

    else:
        # the detector is moving - have to rebin the dataset to contain all values of twoth
        # this is silly but have to set other axis!
        other_axis = (1 if pixel_axis == 0 else 0)
        #other_vector = new_info[other_axis]['values']
        #other_spacing = other_vector[1] - other_vector[0]
        pixels = new_info[pixel_axis]['values']
        twoth = (pixels - qzero_pixel) / pixels_per_degree
        #twoth = twoth[::-1] # reverse
        twoth_min = det_angle_min + twoth.min()
        twoth_max = det_angle_max + twoth.max()
        twoth_max_edge = twoth_max + 1.0 / pixels_per_degree
        dpp = 1.0 / pixels_per_degree
        #output_twoth_bin_edges = arange(twoth_max + dpp, twoth_min - dpp, -dpp)
        output_twoth_bin_edges = arange(twoth_min - dpp, twoth_max + dpp, dpp)
        output_twoth = output_twoth_bin_edges[:-1]
        #other_bin_edges = linspace(other_vector[0], other_vector[-1] + other_spacing, len(other_vector) + 1)
        new_info[pixel_axis]['name'] = 'twotheta' # getting rid of pixel units: substitute twoth
        new_info[pixel_axis]['values'] = output_twoth
        new_info[pixel_axis]['units'] = 'degrees'
        output_shape = [0,0,0]
        output_shape[pixel_axis] = len(output_twoth)
        output_shape[other_axis] = data.shape[other_axis] # len(other_vector)
        output_shape[2] = data.shape[2] # number of columns is unchanged!
        new_data = MetaArray(tuple(output_shape), info=new_info) # create the output data object!

        tth_min = twoth.min()
        tth_max = twoth.max()
        data_in = data.view(ndarray).copy()
        for i, da in enumerate(det_angle):
            twoth_min = da + tth_min
            twoth_max = da + tth_max
            input_twoth_bin_edges = empty(len(pixels) + 1)
            input_twoth_bin_edges[-1] = twoth_max + 1.0 / pixels_per_degree
            input_twoth_bin_edges[:-1] = twoth + da
            #data_cols = ['counts', 'pixels', 'monitor', 'count_time']
            cols = new_info[-2]['cols']
            for col in range(len(cols)):
                input_slice = [slice(None, None), slice(None, None), col]
                #input_slice[pixel_axis] = slice(i, i+1)
                input_slice[other_axis] = i
                array_to_rebin = data_in[tuple(input_slice)]
                new_array = reb.rebin(input_twoth_bin_edges, array_to_rebin, output_twoth_bin_edges)
                new_data.view(ndarray)[tuple(input_slice)] = new_array

    return new_data

@module
def thetaTwothetaToQxQz(data, output_grid, wavelength=5.0, qxmin=-0.003, qxmax=0.003, qxbins=101, qzmin=0.0001, qzmax=0.1, qzbins=101):
    """ Figures out the Qx, Qz values of each datapoint
    and throws them in the correct bin.  If no grid is specified,
    one is created that covers the whole range of Q in the dataset

    If autofill_gaps is True, does reverse lookup to plug holes
    in the output (but pixel count is still zero for these bins)

    **Inputs**

    data (ospec2d): input data

    output_grid (ospec2d?): empty data object with axes defined (optional)

    wavelength (float): override wavelength in data file

    qxmin (float): lower bound of Qx range in rebinning

    qxmax (float): upper bound of Qx range in rebinning

    qxbins (int): number of bins to subdivide the range between qxmin and qxmax

    qzmin (float): lower bound of Qz range in rebinning

    qzmax (float): upper bound of Qz range in rebinning

    qzbins (int): number of bins to subdivide the range between qzmin and qzmax

    **Returns**

    output (ospec2d): output data rebinned into Qx, Qz

    2016-04-01 Brian Maranville
    """
    print("output grid: ", output_grid)
    if output_grid is None:
        info = [{"name": "qx", "units": "inv. Angstroms", "values": linspace(qxmin, qxmax, qxbins) },
            {"name": "qz", "units": "inv. Angstroms", "values": linspace(qzmin, qzmax, qzbins) },]
        old_info = data.infoCopy()
        info.append(old_info[2]) # column information!
        info.append(old_info[3]) # creation story!
        output_grid = MetaArray(zeros((qxbins, qzbins, data.shape[-1])), info=info)
    else:
        outgrid_info = deepcopy(output_grid._info) # take axes and creation story from emptyqxqz...
        outgrid_info[2] = deepcopy(data._info[2]) # take column number and names from dataset
        output_grid = MetaArray(zeros((output_grid.shape[0], output_grid.shape[1], data.shape[2])), info=outgrid_info)

    theta_axis = data._getAxis('theta')
    twotheta_axis = data._getAxis('twotheta')

    qLength = 2.0 * pi / wavelength
    th_array = data.axisValues('theta').copy()
    twotheta_array = data.axisValues('twotheta').copy()

    if theta_axis < twotheta_axis: # then theta is first: add a dimension at the end
        th_array.shape = th_array.shape + (1,)
        twotheta_array.shape = (1,) + twotheta_array.shape
    else:
        twotheta_array.shape = twotheta_array.shape + (1,)
        th_array.shape = (1,) + th_array.shape

    tilt_array = th_array - (twotheta_array / 2.0)
    qxOut = 2.0 * qLength * sin((pi / 180.0) * (twotheta_array / 2.0)) * sin(pi * tilt_array / 180.0)
    qzOut = 2.0 * qLength * sin((pi / 180.0) * (twotheta_array / 2.0)) * cos(pi * tilt_array / 180.0)

    # getting values from output grid:
    outgrid_info = output_grid.infoCopy()
    numcols = len(outgrid_info[2]['cols'])
    qx_array = output_grid.axisValues('qx')
    dqx = qx_array[1] - qx_array[0]
    qz_array = output_grid.axisValues('qz')
    dqz = qz_array[1] - qz_array[0]
    #framed_array = zeros((qz_array.shape[0] + 2, qx_array.shape[0] + 2, numcols))
    target_qx = ((qxOut - qx_array[0]) / dqx).astype(int)
    #return target_qx, qxOut
    target_qz = ((qzOut - qz_array[0]) / dqz).astype(int)

    target_mask = (target_qx >= 0) * (target_qx < qx_array.shape[0])
    target_mask *= (target_qz >= 0) * (target_qz < qz_array.shape[0])
    target_qx_list = target_qx[target_mask]
    target_qz_list = target_qz[target_mask]

    for i, col in enumerate(outgrid_info[2]['cols']):
        values_to_bin = data[:,:,col['name']].view(ndarray)[target_mask]
        outshape = (output_grid.shape[0], output_grid.shape[1])
        hist2d, xedges, yedges = histogram2d(target_qx_list,target_qz_list, \
            bins = (outshape[0],outshape[1]), range=((0,outshape[0]),(0,outshape[1])), weights=values_to_bin)
        output_grid[:,:,col['name']] += hist2d
        #framed_array[target_qz_list, target_qx_list, i] = data[:,:,col['name']][target_mask]

    cols = outgrid_info[2]['cols']
    data_cols = [col['name'] for col in cols if col['name'].startswith('counts')]
    monitor_cols = [col['name'] for col in cols if col['name'].startswith('monitor')]
    # just take the first one...
    if len(monitor_cols) > 0:
        monitor_col = monitor_cols[0]
        data_missing_mask = (output_grid[:,:,monitor_col] == 0)
        for dc in data_cols:
            output_grid[:,:,dc].view(ndarray)[data_missing_mask] = nan;

    #extra_info
    output_grid._info[-1] = data._info[-1].copy()
    print("output shape:", output_grid.shape)
    return output_grid

@module
def thetaTwothetaToAlphaIAlphaF(data):
    """ Figures out the angle in, angle out values of each datapoint
    and throws them in the correct bin.  If no grid is specified,
    one is created that covers the whole range of the dataset

    **Inputs**

    data (ospec2d): input data

    **Returns**

    output (ospec2d): output data rebinned into alpha_i, alpha_f

    2016-04-01 Brian Maranville
    """

    theta_axis = data._getAxis('theta')
    twotheta_axis = data._getAxis('twotheta')

    th_array = data.axisValues('theta').copy()
    twotheta_array = data.axisValues('twotheta').copy()

    two_theta_step = twotheta_array[1] - twotheta_array[0]
    theta_step = th_array[1] - th_array[0]

    af_max = (twotheta_array.max() - th_array.min())
    af_min = (twotheta_array.min() - th_array.max())
    alpha_i = th_array.copy()
    alpha_f = arange(af_min, af_max, two_theta_step)

    info = [{"name": "alpha_i", "units": "degrees", "values": th_array.copy() },
            {"name": "alpha_f", "units": "degrees", "values": alpha_f.copy() },]
    old_info = data.infoCopy()
    info.append(old_info[2]) # column information!
    info.append(old_info[3]) # creation story!
    output_grid = MetaArray(zeros((th_array.shape[0], alpha_f.shape[0], data.shape[-1])), info=info)

    if theta_axis < twotheta_axis: # then theta is first: add a dimension at the end
        alpha_i.shape = alpha_i.shape + (1,)
        ai_out = indices((th_array.shape[0], twotheta_array.shape[0]))[0]
        twotheta_array.shape = (1,) + twotheta_array.shape
    else:
        alpha_i.shape = (1,) + alpha_i.shape
        ai_out = indices((twotheta_array.shape[0], th_array.shape[0]))[1]
        twotheta_array.shape = twotheta_array.shape + (1,)

    af_out = twotheta_array - alpha_i

    # getting values from output grid:
    outgrid_info = output_grid.infoCopy()
    numcols = len(outgrid_info[2]['cols'])
    #target_ai = ((ai_out - th_array[0]) / theta_step).flatten().astype(int).tolist()
    target_ai = ai_out.flatten().astype(int).tolist()
    #return target_qx, qxOut
    target_af = ((af_out - af_min) / two_theta_step).flatten().astype(int).tolist()

    for i, col in enumerate(outgrid_info[2]['cols']):
        values_to_bin = data[:,:,col['name']].view(ndarray).flatten().tolist()
        print(len(target_ai), len(target_af), len(values_to_bin))
        outshape = (output_grid.shape[0], output_grid.shape[1])
        hist2d, xedges, yedges = histogram2d(target_ai, target_af, bins = (outshape[0],outshape[1]), range=((0,outshape[0]),(0,outshape[1])), weights=values_to_bin)
        output_grid[:,:,col['name']] += hist2d

    cols = outgrid_info[2]['cols']
    data_cols = [col['name'] for col in cols if col['name'].startswith('counts')]
    monitor_cols = [col['name'] for col in cols if col['name'].startswith('monitor')]
    # just take the first one...
    if len(monitor_cols) > 0:
        monitor_col = monitor_cols[0]
        data_missing_mask = (output_grid[:,:,monitor_col] == 0)
        for dc in data_cols:
            output_grid[:,:,dc].view(ndarray)[data_missing_mask] = nan

    #extra info changed
    output_grid._info[-1] = data._info[-1].copy()
    return output_grid

@module
def alphaFtoQz(data, wavelength=5.0):
    """ Figures out the Qz values of each datapoint
    and throws them in the correct bin.

    **Inputs**

    data (ospec2d): input data

    wavelength (float): override wavelength in data file

    **Returns**

    qzdata (ospec2d) : output data rebinned into Qz

    2016-04-03 Brian Maranville
    """

    new_info = data.infoCopy()
    # det_angle should be a vector of the same length as the other axis (usually theta)
    # or else just a float, in which case the detector is not moving!
    ndim = len(new_info) - 2 # last two entries in info are for metadata
    ax_name = "alpha_f"
    af_axis = next((i for i in range(len(new_info)-2) if new_info[i]['name'] == ax_name), None)
    if af_axis < 0:
        raise ValueError("error: no %s axis in this dataset" % (ax_name,))

    new_info[af_axis]['name'] = 'Qz'
    af = new_info[af_axis]['values']
    qz = 4.0*pi/wavelength * sin(radians(af))
    new_info[af_axis]['values'] = qz
    new_info[af_axis]['units'] = 'inv. Angstroms'
    new_array = (data.view(ndarray).copy())
    new_data = MetaArray(new_array, info=new_info)
    return new_data

@module
def autogrid(datasets, operation='union', extra_grid_point=True, min_step=1e-10):
    """
    take multiple datasets and create a grid which covers all of them
    - stepsize is smallest stepsize found in datasets
    returns an empty grid with units and labels

    if extra_grid_point is True, adds one point to the end of each axis
    so each dimension is incremented by one (makes edges for rebinning)

    **Inputs**

    datasets (ospec2d[]): input data

    operation (opt:union|intersection): make grid to cover all points (union) or only where overlapped (intersection)

    extra_grid_point (bool): if extra_grid_point is True, adds one point to the end of each axis so each dimension is incremented by one (makes edges for rebinning)

    min_step (float): smallest difference that is not rounded to zero

    **Returns**

    grid (ospec2d): output grid

    2017-05-01 Brian Maranville
    """

    num_datasets = len(datasets)
    dims = 2
    dim_min = zeros((dims, num_datasets))
    dim_max = zeros((dims, num_datasets))
    dim_len = zeros((dims, num_datasets))
    dim_step = zeros((dims, num_datasets))

    for i, data in enumerate(datasets):
        info = data.infoCopy()
        for dim in range(dims):
            av = data.axisValues(dim)
            dim_min[dim, i] = av.min()
            dim_max[dim, i] = av.max()
            dim_len[dim, i] = len(av)
            if dim_len[dim, i] > 1:
                dim_step[dim, i] = float(dim_max[dim, i] - dim_min[dim, i]) / (dim_len[dim, i] - 1)
                # dim0_max[i] += th_step[i] # add back on the last step
            else:
                dim_step[dim, i] = min_step

    final_stepsizes = []
    absolute_max = []
    absolute_min = []
    for dim in range(dims):
        dim_stepsize = dim_step[dim].min()
        if dim_stepsize < min_step:
            dim_stepsize = min_step
        final_stepsizes.append(dim_stepsize)

        if operation == 'union':
            absolute_max.append(dim_max[dim].max())
            absolute_min.append(dim_min[dim].min())
        elif operation == 'intersection':
            absolute_max.append(dim_max[dim].min())
            absolute_min.append(dim_min[dim].max())
        else:
            raise ValueError('operation must be one of "union" or "intersection"')

    # now calculate number of steps:
    output_dims = []
    for dim in range(dims):
        if (dim_len[dim].max() == 1) or (absolute_max[dim] == absolute_min[dim]):
            steps = 1
        else:
            steps = int(round(float(absolute_max[dim] - absolute_min[dim]) / final_stepsizes[dim]))
        if extra_grid_point == True:
            steps += 1
        output_dims.append(steps)

    new_info = datasets[0].infoCopy() # take units etc from first dataset
     # then tack on the number of columns already there:
    output_dims.append(len(new_info[2]['cols']))
    for dim in range(dims):
        new_info[dim]["values"] = (arange(output_dims[dim], dtype='float') * final_stepsizes[dim]) + absolute_min[dim]
    output_grid = MetaArray(zeros(tuple(output_dims)), info=new_info)
    return output_grid


@module
def combine(datasets, grid, operation="union"):
    """
    Combine multiple datasets with or without overlap, adding
    all the values in the time, monitor and data columns, and populating
    the pixels column (number of pixels in each new bin)

    **Inputs**

    datasets (ospec2d[]): input data

    grid (ospec2d?): optional grid

    operation (opt:union|intersection): make grid to cover all points (union) or only where overlapped (intersection) for autogridding

    **Returns**

    combined (ospec2d) : datasets added together

    2017-05-01 Brian Maranville
    """

    if grid is None:
        grid = autogrid(datasets, operation=operation)
    for dataset in datasets:
        grid = add_to_grid(dataset, grid)
    # extra info changed
    # strip info that is meaningless in combined dataset: (filename, start_time, end_time)
    for key in ['filename', 'start_datetime', 'end_datetime']:
        if key in grid._info[-1]: grid._info[-1].pop(key)
    return grid

def subtract(minuend, subtrahend):
    """ 
    Takes two data objects and subtracts them.
    If no grid is provided, use Autogrid filter to generate one.

    **Inputs**

    minuend (ospec2d): input data

    subtrahend (ospec2d): 

    grid (ospec2d?): optional grid

    operation (opt:union|intersection): make grid to cover all points (union) or only where overlapped (intersection) for autogridding

    **Returns**

    combined (ospec2d) : datasets added together
    """
    #subtrahend = subtrahend[0] # can only subtract one thing... but from many.
    if len(minuend) == len(subtrahend): pass # go with it.
    elif len(subtrahend) == 1: subtrahend = [subtrahend[0] for m in minuend] # broadcast
    else: raise Exception("I don't know what to do with unmatched argument lengths")
    results = []
    for m, s in zip(minuend, subtrahend):
        dim_m = len(m.shape) - 1
        dim_s = len(s.shape) - 1
        if dim_m == 2 and dim_s == 1:
            print("subtract vector from matrix (broadcast subtrahend)")
            s_units = s._info[0]['units']
            m1_units = m._info[0]['units']
            m2_units = m._info[1]['units']
            if s_units == m1_units: active_axis = 0
            elif s_units == m2_units: active_axis = 1
            else: raise Exception("no matching units to subtract from") # bail out!

            new_axisvals = [m._info[0]['values'].copy(), m._info[1]['values'].copy()]
            update_axisvals = new_axisvals[active_axis]
            s_axisvals = s._info[0]['values']
            overlap = slice(get_index(update_axisvals, s_axisvals[0]), get_index(update_axisvals, s_axisvals[-1]))
            print(overlap)
            new_axisvals[active_axis] = update_axisvals[overlap]
            full_overlap = [slice(None, None), slice(None, None)]
            full_overlap[active_axis] = overlap
            full_overlap = tuple(full_overlap)

            output_array = []

            dims = 2
            data_edges = []
            for dim in range(dims):
                av = new_axisvals[dim].copy()
                dspacing = (av.max() - av.min()) / (len(av) - 1)
                edges = resize(av, len(av) + 1)
                edges[-1] = av[-1] + dspacing
                data_edges.append(edges)



            #bin_edges = [[data_edges[0][0], data_edges[0][-1]],[data_edges[1][0], data_edges[1][-1]]]
            av = s_axisvals #give the same treatment to the subtrahend asixvals
            dspacing = (av.max() - av.min()) / (len(av) - 1)
            edges = resize(av, len(av) + 1)
            edges[-1] = av[-1] + dspacing
            #bin_edges[active_axis] = edges

            #new_s = reb.rebin(edges, s['Measurements':col], data_edges[active_axis])

            #new_sshape = [1,1]
            #new_sshape[active_axis] = len(new_saxisvals)
            #new_saxisvals.shape = tuple(new_sshape)

            #new_array = reb.rebin2d(data_edges[0], data_edges[1], array_to_rebin, bin_edges[0], bin_edges[1])

            print(m._info[0]['units'], m._info[1]['units'], s._info[0]['units'])
            data_array = m.view(ndarray).copy()[full_overlap]
            new_info = m.infoCopy()
            new_info[0]['values'] = new_axisvals[0]
            new_info[1]['values'] = new_axisvals[1]
            print(data_array.shape, new_axisvals[0].shape, new_axisvals[1].shape)
            new_data = MetaArray(data_array, info=new_info)
            subtractable_columns = [c['name'] for c in s._info[1]['cols'] if c['name'].startswith('counts')]
            #subtractable_columns = dict(subtractable_columns)
            print("subtractable columns:", subtractable_columns)
            for i, col in enumerate(new_info[2]['cols']):
                if col['name'].startswith('counts') and col['name'] in subtractable_columns:
                    new_s = reb.rebin(edges, s['Measurements':col['name']], data_edges[active_axis])
                    new_sshape = [1, 1]
                    new_sshape[active_axis] = len(new_s)
                    new_s.shape = tuple(new_sshape)
                    new_data['Measurements':col['name']] -= new_s
            results.append(new_data)
        elif dim_m == 2 and dim_s == 2:
            print("subtract matrix from matrix (in overlap)")
            grid = Autogrid().apply([m, s], operation='intersection')
            grid = self.add_to_grid(m, grid, counts_multiplier=1.0)
            grid = self.add_to_grid(s, grid, counts_multiplier=-1.0)
            print(grid._info[0]['units'], m._info[1]['units'], s._info[0]['units'], s._info[1]['units'])
            results.append(grid)
        elif dim_m == 1 and dim_s == 2:
            print("can't do this.")
            print(m._info[0]['units'], s._info[0]['units'], s._info[1]['units'])
            results.append(m)
        elif dim_m == 1 and dim_s == 1:
            print("subtract vector from vector (in overlap)")
            print(m._info[0]['units'], s._info[0]['units'])
            results.append(m)

    return results


def add_to_grid(dataset, grid, counts_multiplier=1.0):
    dims = 2
    grid_slice = [slice(None, None, 1),] * dims
    bin_edges = []
    for dim in range(dims):
        av = grid.axisValues(dim).copy()
        dspacing = (av[-1] - av[0]) / (len(av) - 1)
        edges = resize(av, len(av) + 1)
        edges[-1] = av[-1] + dspacing
        if dspacing < 0:
            edges = edges[::-1] # reverse
            grid_slice[dim] = slice(None, None, -1)
        bin_edges.append(edges)

    data_edges = []
    data_slice = [slice(None, None, 1),] * dims
    for dim in range(dims):
        av = dataset.axisValues(dim).copy()
        dspacing = (av[-1] - av[0]) / (len(av) - 1)
        edges = resize(av, len(av) + 1)
        edges[-1] = av[-1] + dspacing
        if dspacing < 0:
            edges = edges[::-1] # reverse
            data_slice[dim] = slice(None, None, -1)
        data_edges.append(edges)

    data_slice = tuple(data_slice)
    grid_slice = tuple(grid_slice)
    new_info = dataset.infoCopy()
    for i, col in enumerate(new_info[2]['cols']):
        #if col['name'] in cols_to_add:
        if 'counts' in col['name']:
            multiplier = counts_multiplier
        else:
            multiplier = 1.0  # add monitor counts and time always
        array_to_rebin = dataset[:, :, col['name']].view(ndarray)
        #print(data_edges, bin_edges)
        new_array = reb.rebin2d(data_edges[0], data_edges[1], array_to_rebin[data_slice], bin_edges[0], bin_edges[1])
        grid[:, :, col['name']] += (multiplier * new_array[grid_slice])

    return grid

#################
# Loader stuff
#################
DETECTOR_ACTIVE = (320, 340)

@module
def LoadMAGIKPSDMany(fileinfo=None, collapse=True, collapse_axis='y', auto_PolState=False, PolState='', flip=True, transpose=True):
    """
    loads a data file into a MetaArray and returns that.
    Checks to see if data being loaded is 2D; if not, quits

    Need to rebin and regrid if the detector is moving...

    **Inputs**

    fileinfo (fileinfo[]): Files to open.

    collapse {Collapse along one of the axes} (bool): sum over axis of detector

    collapse_axis {number index of axis to collapse along} (opt:x|y): axis to sum over

    auto_PolState {Automatic polarization identify} (bool): automatically determine the polarization state from entry name

    PolState (str): polarization state if not automatically detected

    flip (bool): flip the data up and down

    transpose (bool): transpose the data

    **Returns**

    output (ospec2d[]): all the entries loaded.

    2016-04-04 Brian Maranville
    """
    outputs = []
    kwconfig = {
        "collapse": collapse,
        "collapse_axis": collapse_axis,
        "auto_PolState": auto_PolState,
        "PolState": PolState,
        "flip": flip,
        "transpose": transpose
    }
    for fi in fileinfo:
        template_def = {
          "name": "loader_template",
          "description": "Offspecular remote loader",
          "modules": [
            {"module": "ncnr.ospec.LoadMAGIKPSD", "version": "0.1", "config": {}}
          ],
          "wires": [],
          "instrument": "ncnr.magik",
          "version": "0.0"
        }
        template = Template(**template_def)
        config = {"0": {"fileinfo": {"path": fi['path'], "source": fi['source'], "mtime": fi['mtime']}}}
        config["0"].update(kwconfig)
        nodenum = 0
        terminal_id = "output"

        retval = process_template(template, config, target=(nodenum, terminal_id))
        outputs.extend(retval.values)
    return outputs

@cache
@module
def LoadMAGIKPSD(fileinfo=None, collapse=True, collapse_axis='y', auto_PolState=False, PolState='', flip=True, transpose=True):
    """
    loads a data file into a MetaArray and returns that.
    Checks to see if data being loaded is 2D; if not, quits

    Need to rebin and regrid if the detector is moving...

    **Inputs**

    fileinfo (fileinfo): File to open.

    collapse {Collapse along one of the axes} (bool): sum over axis of detector

    collapse_axis {number index of axis to collapse along} (opt:x|y): axis to sum over

    auto_PolState {Automatic polarization identify} (bool): automatically determine the polarization state from entry name

    PolState (str): polarization state if not automatically detected

    flip (bool): flip the data up and down

    transpose (bool): transpose the data

    **Returns**

    output (ospec2d[]): all the entries loaded.

    2016-04-04 Brian Maranville
    """

    path, mtime, entries = fileinfo['path'], fileinfo['mtime'], fileinfo['entries']
    name = basename(path)
    fid = BytesIO(url_get(fileinfo))
    file_obj = h5_open_zip(name, fid)
    return loadMAGIKPSD_helper(file_obj, name, path, collapse=collapse, collapse_axis=collapse_axis, auto_PolState=auto_PolState, PolState=PolState, flip=flip, transpose=transpose)

def LoadMAGIKPSDFile(path, **kw):
    return loadMAGIKPSD_helper(h5_open_zip(path), path, path, **kw)

def loadMAGIKPSD_helper(file_obj, name, path, collapse=True, collapse_axis='y', auto_PolState=False, PolState='', flip=True, transpose=True) -> List[MetaArray]:
    lookup = {"DOWN_DOWN":"_down_down", "UP_DOWN":"_up_down", "DOWN_UP":"_down_up", "UP_UP":"_up_up", "entry": ""}
    #nx_entries = LoadMAGIKPSD.load_entries(name, fid, entries=entries)
    #fid.close()

    #if not (len(file_obj.detector.counts.shape) == 2):
        # not a 2D object!
    #    return
    for entryname, entry in file_obj.items():
        active_slice = slice(None, DETECTOR_ACTIVE[0], DETECTOR_ACTIVE[1])
        counts_value = entry['DAS_logs/areaDetector/counts'][:, 1:DETECTOR_ACTIVE[0]+1, :DETECTOR_ACTIVE[1]]
        dims = counts_value.shape
        ndims = len(dims)
        if auto_PolState:
            PolState = lookup.get(entryname, "")
        # force PolState to a regularized version:
        if not PolState in lookup.values():
            PolState = ''
        #datalen = file_obj.detector.counts.shape[0]
        if ndims == 2:
            if DEBUG: print("2d")
            ypixels = dims[0]
            xpixels = dims[1]
        elif ndims >= 3:
            if DEBUG: print("3d")
            frames = dims[0]
            xpixels = dims[1]
            ypixels = dims[2]


        # doesn't really matter; changing so that each keyword (whether it took the default value
        # provided or not) will be defined
        #    if not PolState == '':
        #        creation_story += ", PolState='{0}'".format(PolState)
        # creation_story += ")"


        if ndims == 2: # one of the dimensions has been collapsed.
            info = []
            info.append({"name": "xpixel", "units": "pixels", "values": arange(xpixels) }) # reverse order
            samp_angle = entry['DAS_logs/sampleAngle/softPosition'][()]
            det_angle = entry['DAS_logs/detectorAngle/softPosition'][()]
            if samp_angle.size > 1:
                yaxis = entry['DAS_logs/sampleAngle/softPosition']
                yaxisname = "theta"
            elif det_angle.size > 1:
                yaxis = entry['DAS_logs/detectorAngle/softPosition']
                yaxisname = "det_angle"
            else:
                # need to find the one that's moving...
                yaxis = entry['data/x']
                yaxisname = yaxis.path
            yaxisunits = yaxis.attrs['units']
            yaxisvalues = yaxis[()]
            info.append({"name": yaxisname, "units": yaxisunits, "values": yaxisvalues})
            info.extend([
                    {"name": "Measurements", "cols": [
                            {"name": "counts"},
                            {"name": "pixels"},
                            {"name": "monitor"},
                            {"name": "count_time"}]},
                    {"PolState": PolState, "filename": name, "start_datetime": entry['start_time'][0], "friendly_name": entry['DAS_logs/sample/name'][0],
                     "entry": entryname, "path":path, "det_angle":entry['DAS_logs/detectorAngle/softPosition'][()],
                     "theta": entry['DAS_logs/sampleAngle/softPosition'][()]}]
                )
            data_array = zeros((xpixels, ypixels, 4))
            mon =  entry['DAS_logs']['counter']['liveMonitor'][()]
            count_time = entry['DAS_logs']['counter']['liveTime'][()]
            if ndims == 2:
                mon.shape = (1,) + mon.shape # broadcast the monitor over the other dimension
                count_time.shape = (1,) + count_time.shape
            counts = counts_value
            if transpose == True: counts = counts.swapaxes(0,1)
            if flip == True: counts = flipud(counts)
            data_array[..., 0] = counts
            #data_array[..., 0] = file_obj.detector.counts
            data_array[..., 1] = 1
            data_array[..., 2] = mon
            data_array[..., 3] = count_time
            # data_array[:,:,4]... I wish!!!  Have to do by hand.
            data = MetaArray(data_array, dtype='float', info=info)
            data.friendly_name = name # goes away on dumps/loads... just for initial object.
            ouput = [data]

        elif ndims == 3: # then it's an unsummed collection of detector shots.  Should be one sample and detector angle per frame
            if collapse == True:
                info = []
                xaxis = "xpixel" if collapse_axis == 'y' else "ypixel"
                xdim = xpixels if collapse_axis == 'y' else ypixels
                xaxisvalues = arange(xdim)
                info.append({"name": xaxis, "units": "pixels", "values": xaxisvalues }) # reverse order
                samp_angle = entry['DAS_logs/sampleAngle/softPosition'][()]
                det_angle = entry['DAS_logs/detectorAngle/softPosition'][()]
                if samp_angle.size > 1:
                    yaxis = entry['DAS_logs/sampleAngle/softPosition']
                    yaxisname = "theta"
                elif det_angle.size > 1:
                    yaxis = entry['DAS_logs/detectorAngle/softPosition']
                    yaxisname = "det_angle"
                else:
                    # need to find the one that's moving...
                    yaxis = entry['data/x']
                    yaxisname = yaxis.path
                yaxisunits = yaxis.attrs['units']
                yaxisvalues = yaxis[()]
                info.append({"name": yaxisname, "units": yaxisunits, "values": yaxisvalues})
                info.extend([
                        {"name": "Measurements", "cols": [
                                {"name": "counts"},
                                {"name": "pixels"},
                                {"name": "monitor"},
                                {"name": "count_time"}]},
                        {"PolState": PolState, "start_datetime": entry['start_time'][0], "path":path,
                         "det_angle": det_angle.tolist(),
                         "theta": samp_angle.tolist(),
                         "filename": name,
                         "friendly_name": entry['DAS_logs/sample/name'][0], "entry": entryname}]
                    )
                data_array = zeros((xdim, frames, 4))
                mon =  entry['DAS_logs']['counter']['liveMonitor'][()]
                count_time = entry['DAS_logs']['counter']['liveTime'][()]
                if ndims == 3:
                    mon.shape = (1,) + mon.shape # broadcast the monitor over the other dimension
                    count_time.shape = (1,) + count_time.shape
                axis_to_sum = 2 if collapse_axis == 'y' else 1
                counts = sum(counts_value, axis=axis_to_sum)
                if transpose == True: counts = counts.swapaxes(0,1)
                if flip == True: counts = flipud(counts)
                data_array[..., 0] = counts
                #data_array[..., 0] = file_obj.detector.counts
                data_array[..., 1] = 1
                data_array[..., 2] = mon
                data_array[..., 3] = count_time
                # data_array[:,:,4]... I wish!!!  Have to do by hand.
                data = MetaArray(data_array, dtype='float', info=info)
                data.friendly_name = name # goes away on dumps/loads... just for initial object.
                output = [data]
            else: # make separate frames
                infos = []
                data = []
                samp_angle =  entry['DAS_logs/sampleAngle/softPosition'][()].astype('float')
                if samp_angle.shape[0] == 1:
                    samp_angle = ones((frames,)) * samp_angle
                det_angle = entry['DAS_logs/detectorAngle/softPosition'][()].astype('float')
                if det_angle.shape[0] == 1:
                    det_angle = ones((frames,)) * det_angle
                count_time = entry['DAS_logs/counter/liveTime'][()]
                if count_time.shape[0] == 1:
                    count_time = ones((frames,)) * count_time
                mon =  entry['DAS_logs/counter/liveMonitor'][()]
                if mon.shape[0] == 1:
                    mon = ones((frames,)) * mon
                for i in range(frames):
                    info = []
                    info.append({"name": "xpixel", "units": "pixels", "values": list(range(xpixels)) })
                    info.append({"name": "ypixel", "units": "pixels", "values": list(range(ypixels)) })
                    info.extend([
                        {"name": "Measurements", "cols": [
                                {"name": "counts"},
                                {"name": "pixels"},
                                {"name": "monitor"},
                                {"name": "count_time"}]},
                        {"PolState": PolState, "filename": name, "start_datetime": entry['start_time'][0], "friendly_name": entry['DAS_logs/sample/name'][0],
                         "entry": entryname, "path":path, "samp_angle": samp_angle[i], "det_angle": det_angle[i]}]
                    )
                    data_array = zeros((xpixels, ypixels, 4))
                    counts = counts_value[i]
                    if flip == True: counts = flipud(counts)
                    data_array[..., 0] = counts
                    data_array[..., 1] = 1
                    data_array[..., 2] = mon[i]
                    data_array[..., 3] = count_time[i]
                    # data_array[:,:,4]... I wish!!!  Have to do by hand.
                    subdata = MetaArray(data_array, dtype='float', info=info)
                    subdata.friendly_name = name + ("_%d" % i) # goes away on dumps/loads... just for initial object.
                    data.append(subdata)
                    output = data
    return output

def demo():
    from reductus.dataflow import fetch
    fetch.DATA_SOURCES = [{"name": "ncnr", "url": "https://ncnr.nist.gov/pub/"}]
    fileinfo = {
        'mtime': 1457795231.0,
        'path': 'ncnrdata/cgd/201603/21237/data/wp10v132.nxz.cgd',
        'source': 'ncnr',
        'entries': ['entry']
    }
    return LoadMAGIKPSD(fileinfo)

def test():
    loaded_files = demo()
    assert len(loaded_files) == 1
    loaded_file = loaded_files[0]
    assert loaded_file.get_metadata()['filename'] == 'wp10v132.nxz.cgd'


from __future__ import print_function

import numpy as np

from reductus.dataflow.lib import err1d

def extend(a, b):
    """
    Extend *a* to match the number of dimensions of *b*.

    This adds dimensions to the end of *a* rather than the beginning. It is
    equivalent to *a[..., None, None]* with the right number of None elements
    to make the number of dimensions match (or np.newaxis if you prefer).

    For example::

        from numpy.random import rand
        a, b = rand(3, 4), rand(3, 4, 2)
        a + b
        ==> ValueError: operands could not be broadcast together with shapes (3,4) (3,4,2)
        c = extend(a, b) + b
        c.shape
        ==> (3, 4, 2)

    Numpy broadcasting rules automatically extend arrays to the beginning,
    so the corresponding *lextend* function is not needed::

        c = rand(3, 4) + rand(2, 3, 4)
        c.shape
        ==> (2, 3, 4)
    """
    if np.isscalar(a):
        return a
    extra_dims = (np.newaxis,)*(b.ndim-a.ndim)
    return a[(..., *extra_dims)]

def indent(text, prefix="  "):
    """
    Add a prefix to every line in a string.
    """
    return "\n".join(prefix+line for line in text.splitlines())


def group_data(datasets):
    """
    Groups data files by intent and polarization cross section.

    Returns a dictionary with the groups, keyed by (intent,polarization).
    """
    # TODO: also need to group by temperature/field
    groups = {}
    for d in datasets:
        groups.setdefault((d.intent, d.polarization), []).append(d)
    return groups


def group_by_xs(datasets):
    """
    Return datasets grouped by polarization cross section.
    """
    cross_sections = {}
    for data in datasets:
        cross_sections.setdefault(data.polarization, []).append(data)

    #print("datasets", [":".join((d.name, d.entry, d.polarization, d.intent)) for d in datasets])
    #print("xs", cross_sections)
    return cross_sections

def group_by_key(key, datasets):
    """
    Return datasets grouped by a value that can be found in a refldata file.
    Handle dotted namespace through recursive lookup.
    Handle union with comma. (e.g. key = "polarization,sample.name" would
    create group where sample.name and polarization are the same for all)
    """
    groups = {}
    key_items = key.split(",")
    for data in datasets:
        groupkey = []
        for item in key_items:
            item = item.strip()
            value = data
            for k in item.split("."):
                value = getattr(value, k)
            groupkey.append(value)
        groupkey = tuple(sorted(groupkey))
        groups.setdefault(groupkey, []).append(data)
    return groups


def group_by_intent(datasets):
    """
    Return datasets grouped by intent.
    """
    intents = {}
    for data in datasets:
        intents.setdefault(data.intent, []).append(data)

    #print("datasets", [":".join((d.name, d.entry, d.polarization, d.intent)) for d in datasets])
    #print("xs", cross_sections)
    return intents


def nearest(x, xp, fp=None):
    """
    Return the *fp* value corresponding to the *xp* that is nearest to *x*.

    If *fp* is missing, return the index of the nearest value.
    """
    if len(xp) == 1:
        if np.isscalar(x):
            return fp[0] if fp is not None else 0
        else:
            return np.array(len(x)*(fp if fp is not None else [0]))

    # if fp is not provided, want to return f as an index into the array xp
    # for the target values x, so set it to integer indices.  if fp is
    # provided, make sure it is an array.
    fp = np.arange(len(xp)) if fp is None else np.asarray(fp)


    # make sure that the xp array is sorted
    xp = np.asarray(xp)
    if np.any(np.diff(xp) < 0.):
        index = np.argsort(xp)
        xp, fp = xp[index], fp[index]

    # find the midpoints of xp and use that as the index
    xp = 0.5*(xp[:-1] + xp[1:])
    return fp[np.searchsorted(xp, x)]


def plot_sa(data):
    """
    Plot spin asymmetry data.
    """
    from matplotlib import pyplot as plt
    from uncertainties.unumpy import uarray as U, nominal_values, std_devs
    from reductus.dataflow.lib.errutil import interp
    # TODO: interp doesn't test for matching resolution
    data = dict((d.polarization, d) for d in data)
    pp, mm = data['++'], data['--']
    v_pp = U(pp.v, pp.dv)
    v_mm = interp(pp.x, mm.x, U(mm.v, mm.dv))
    sa = (v_pp - v_mm) / (v_pp + v_mm)
    v, dv = nominal_values(sa), std_devs(sa)
    plt.errorbar(pp.x, v, yerr=dv, fmt='.', label=pp.name)
    plt.xlabel("%s (%s)"%(pp.xlabel, pp.xunits) if pp.xunits else pp.xlabel)
    plt.ylabel(r'$(R^{++} -\, R^{--}) / (R^{++} +\, R^{--})$')


def test_nearest():
    # length 1 arrays
    xp, fp = [1], [5]
    assert nearest(0, xp) == 0
    assert (nearest([0], xp) == [0]).all()
    assert (nearest([0, 1], xp) == [0, 0]).all()
    assert nearest(0, xp, fp) == fp[0]
    assert (nearest([0], xp, fp) == [fp[0]]).all()
    assert (nearest([0, 1], xp, fp) == [fp[0]]*2).all()

    # constants as arrays
    xp, fp = [1, 1, 1], [5, 5, 5]
    assert nearest(0, xp) == 0
    assert (nearest([0], xp) == [0]).all()
    assert (nearest([0, 1], xp) == [0, 0]).all()
    assert nearest(0, xp, fp) == fp[0]
    assert (nearest([0], xp, fp) == [fp[0]]).all()
    assert (nearest([0, 1], xp, fp) == [fp[0]]*2).all()

    # actual arrays
    xp, fp = [1, 2, 3], [4, 5, 6]
    assert nearest(0, xp) == 0
    assert (nearest([0], xp) == [0]).all()
    assert (nearest([0, 1, 1.1, 1.6, 2.1, 2.9, 3, 3.1], xp)
            == [0, 0, 0, 1, 1, 2, 2, 2]).all()
    assert nearest(0, xp, fp) == fp[0]
    assert (nearest([0], xp, fp) == [fp[0]]).all()
    assert (nearest([0, 1, 1.1, 1.6, 2.1, 2.9, 3, 3.1], xp, fp)
            == [fp[i] for i in [0, 0, 0, 1, 1, 2, 2, 2]]).all()

    # unsorted arrays
    xp, fp = [1, 3, 2], [4, 5, 6]
    assert nearest(0, xp) == 0
    assert (nearest([0], xp) == [0]).all()
    assert (nearest([0, 1, 1.1, 1.6, 2.1, 2.9, 3, 3.1], xp)
            == [0, 0, 0, 2, 2, 1, 1, 1]).all()
    assert nearest(0, xp, fp) == fp[0]
    assert (nearest([0], xp, fp) == [fp[0]]).all()
    assert (nearest([0, 1, 1.1, 1.6, 2.1, 2.9, 3, 3.1], xp, fp)
            == [fp[i] for i in [0, 0, 0, 2, 2, 1, 1, 1]]).all()


def poisson_average(y, dy, norm='monitor'):
    r"""
    Return the Poisson average of a rate vector *y +/- dy*.

    If y, dy is multidimensional then average the first dimension, returning
    an item of one fewer dimentsions.

    Use *norm='monitor'* When counting against monitor (the default) or
    *norm='time'* when counting against time.  Use *norm='none'* if *y, dy*
    is unnormalized, and the poisson sum should be returned. Use *norm='gauss'*
    for the simple weighted gaussian average.

    The count rate is expressed as the number of counts in an interval $N$
    divided by the interval $M$.  The rate for the combined interval should
    match the rate you would get if you counted for the entire interval,
    which is $\sum N_i / \sum M_i$.  We do this by inferring the counts
    and intervals from the rate and uncertainty, adding them together, and
    dividing to get the average rate over the entire interval.

    With counts $N$ and monitors $M$ both from Poisson distributions, the
    uncertainties are $\sqrt N$ and $\sqrt M$ respectively, and gaussian
    error propagation gives

    .. math::
       :nowrap:

        \begin{eqnarray}
        y &=& N/M \\
        \left(\frac{\Delta y}{y}\right)^2
            &=& \left(\frac{\Delta N}{N}\right)^2
              + \left(\frac{\Delta M}{M}\right)^2 \\
            &=& \left(\frac{1}{N}\right)
              + \left(\frac{1}{M}\right) \\
        \Delta y &=& y \sqrt{1/N + 1/M}
        \end{eqnarray}

    Inverting, we get

    .. math::
       :nowrap:

        \begin{eqnarray}
        M &=& y (y+1) / \Delta y^2 \\
        N &=& y M
        \end{eqnarray}

    When counts are zero, $y = 0$ and $M = 0$ according to the above formula.
    To correctly average intervals that may include zero counts, we need to
    be sure that the count uncertainty $\Delta N = 1$ when $N = 0$, which
    leads to a count rate of $0 \pm 1/M$, and the formula above needs to be
    adjusted accordingly, with $M = 1/ \Delta y$ when $N = 0$.

    To average a group of measurements $y_1, \ldots, y_n$ we first
    convert to counts and monitors, then set the rate to the sum of
    the counts over the sum of the monitors.  This gives

    .. math::
       :nowrap:

        \begin{eqnarray}
        M_i &=& y_i (y_i + 1) / \Delta y_i^2 \\
        N_i &=& y_i M_i \\
        \bar y &=& \sum N_i / \sum M_i \\
        \Delta \bar y &=& \bar y \sqrt{1/\sum_{N_i} + 1/\sum_{M_i}}
        \end{eqnarray}

    When counting against time the monitor uncertainty $\Delta M$ is
    effectively zero compared to uncertainty in the counts, and so
    the formulas are a little simpler:

    .. math::

        y &= N/M \\
        \Delta y &= y \sqrt{1/N} \\
        M &= y / \Delta y^2 \\
        N &= y M

    Again, zero counts leads to $M = 1 / \Delta y$.

    Averaging gives

    .. math::
       :nowrap:

        \begin{eqnarray}
        M_i &=& y_i / \Delta y_i^2 \\
        N_i &=& y_i M_i \\
        \bar y &=& \sum N_i / \sum M_i \\
        \Delta \bar y &=& \bar y \sqrt{1/\sum N_i}
        \end{eqnarray}

    This algorithm is robust against scale factors, such as detector
    efficiency, attenuators and dead-time correction. Assume the underlying
    rate $r$ for the averaged points is identical (otherwise why are you mixing
    them?) with an independent scale factor $C_i$ applied to each point. Then

    .. math::
        :nowrap:

        \begin{eqnarray}
        y_i' &=& C_i N_i/M_i \\
        \Delta y_i' &=& C_i \sqrt{N_i}/M_i \\
        M_i' &=& y_i' / \Delta y_i^2' = M_i/C_i \\
        N_i' &=& y_i' M_i' = N_i
        \end{equnarray}

    and

    .. math::
        :nowrap:

        \begin{eqnarray}
        r  &\approx& C_i N_i / M_i \Rightarrow M_i/C_i \approx N_i/r \\
        M^{+} &=& \sum M_i' \approx \sum N_i / r \\
        N^{+} &=& \sum N_i' = \sum N_i \\
        \bar y' &=& N^{+} / M^{+} \approx r \\
        \Delta \bar y' = \sqrt{\bar y' / M^{+}} = \sqrt{N^{+}} / M^{+} \approx r / \sqrt{\sum N_i}

    That is, $\bar y', \Delta \bar y'$ will be computed from the scaled data,
    again with uncertainty proportional to $\sqrt{N}$.

    Monte Carlo experiments confirm that this holds in practice, giving
    reasonable values for the signal when subtracting mixed background from
    mixed signal+background, even when signal rate and background rate vary
    amongst the points measured. This is not true for Gaussian averaging, which
    fails when the counts are approximately less than 100.
    """
    if norm not in ("monitor", "time", "gauss", "none"):
        raise ValueError("expected norm to be time, monitor or none")

    # Check whether we are combining rates or counts.  If it is counts,
    # then simply sum them, and sum the uncertainty in quadrature. This
    # gives the expected result for poisson statistics, with counts over
    # the combined interval having variance equal to the sum of the counts.
    # It even gives correct results for very low count rates with many of
    # the individual counts giving zero, so long as variance on zero counts
    # is set to zero rather than one.
    if norm == "none":
        bar_y = np.sum(y, axis=0)
        bar_dy = np.sqrt(np.sum(dy**2, axis=0))
        return bar_y, bar_dy

    dy = dy + (dy == 0)  # Protect against zero counts in division
    if norm == "gauss":
        Swx = np.sum(y/dy**2, axis=0)
        Sw = np.sum(dy**-2, axis=0)
        bar_y = Swx / Sw
        bar_dy = 1/np.sqrt(Sw)
        return bar_y, bar_dy

    # Recover monitor and counts
    monitors = y*(y+1)/dy**2 if norm == "monitor" else y/dy**2 # if "time"
    monitors[y == 0] = 1./dy[y == 0]  # Special handling for 0 counts
    counts = y*monitors

    # Compute average rate
    combined_monitors = np.sum(monitors, axis=0)
    combined_counts = np.sum(counts, axis=0)
    bar_y = combined_counts/combined_monitors
    if norm == "time":
        bar_dy = np.sqrt(bar_y/combined_monitors)
    elif np.isscalar(bar_y):
        if bar_y == 0:
            # When bar_y is zero then 1/N is undefined and so sqrt(1/N + 1/M) fails.
            # Instead use |dy| = 1/M*sqrt((dN)^2 + 1/M) with dN = 1.
            bar_dy = 1./combined_monitors * np.sqrt(1. + 1./combined_monitors)
        else:
            # When bar_y is not zero then use |dy| = N/M * sqrt(1/N + 1/M)
            bar_dy = bar_y * np.sqrt(1./combined_counts + 1./combined_monitors)
    else:
        # Following the scalar case above, first build bar_dy assuming
        # that y is zero since it works for all y, then fill in the values
        # for y not zero. Can't do this the other way since the expression
        # for y not zero will raise errors when y is zero.
        bar_dy = 1./combined_monitors * np.sqrt(1. + 1./combined_monitors)
        idx = (bar_y != 0)
        bar_dy[idx] = bar_y[idx] * np.sqrt(1./combined_counts[idx]
                                           + 1./combined_monitors[idx])

    #print("est. monitors:", monitors)
    #print("est. counts:", counts)
    #print("poisson avg", counts.shape, bar_y.shape, bar_dy.shape)
    return bar_y, bar_dy


def gaussian_average(y, dy, w, dw=0):
    bar_y, bar_y_var = err1d.average(y, dy**2, w, dw**2)
    return bar_y, np.sqrt(bar_y_var)

def demo_sub():
    """
    Check subtraction with mixed background and mixed scale.

    The assumption that $r$ is fixed does not hold when binning $Q$ for
    Candor. Both the signal (due to varying resolution) and the background
    due to measurement geometry, detector properties, etc. vary from point
    to point, and so the simplification that $M^{+} \approx \sum N_i / r$
    no longer holds.

    Instead we want the resulting signal to be the weighted average of the
    different signals when mixing signal+background and subtracting the
    mixed background.
    """
    from numpy import mean, std, sqrt
    from numpy.random import poisson

    NP = 5e-6  # 5 us non-paralyzing dead time
    def deadtime(r):
        return 1/(1 + r*NP)

    def _sim_one(spec, back, slit, spec_time, back_time):
        s = (spec+back)*slit
        b = back*slit
        S = poisson(s*deadtime(s)*spec_time)
        B = poisson(b*deadtime(b)*back_time)
        Sscale = spec_time*slit*deadtime(s)
        Bscale = back_time*slit*deadtime(b)
        # Be sure to account for uncertainty on zero counts.
        Sr, dSr = S/Sscale, np.sqrt(S+(S==0))/Sscale
        Br, dBr = B/Bscale, np.sqrt(B+(B==0))/Bscale
        Sm, dSm = poisson_average(Sr, dSr, norm='time')
        Bm, dBm = poisson_average(Br, dBr, norm='time')
        sub = Sm - Bm
        dsub = np.sqrt(dSm**2 + dBm**2)
        #print(s, S, Sscale, Sr, Sm)
        #print(b, B, Bscale, Br, Bm)
        #print(sub, dsub)
        return sub, dsub

    def _sim(spec, back, slit, t=100, n=10000):
        spec_time, back_time = np.full_like(spec, t), np.full_like(spec, t)
        target = np.sum(spec*slit)/np.sum(slit)
        out = [_sim_one(spec, back, slit, spec_time, back_time)
               for _ in range(n)]
        out = np.asarray(out)
        print("target %.4g r %.4g +/- %.2g, dr %.2g"
              % (target, mean(out[:, 0]), std(out[:, 0]), sqrt(mean(out[:, 1]**2))))

    k = 40
    spec = np.random.rand(k)*1e-7 + 1e-6
    back = np.random.rand(k)*1e-6
    slit = np.linspace(1e6, 4e6, k)

    if 0:
        spec = np.array([1e-6, 1.1e-6, 0.9e-6])  # a little variation due to resolution
        back = np.array([1e-8, 1e-6, 1e-7]) # high variation to see what happens
        slit = np.array([1e6, 2e6, 3e6]) # a lot of variation due to spectrum

    point_time = 3
    #print("for spec", spec, "back", back)
    _sim(spec, back, slit, n=1000, t=point_time)
    print("back /10")
    _sim(spec, back/100, slit, n=1000, t=point_time)
    print("back x100")
    _sim(spec, back*100, slit, n=1000, t=point_time)

def demo_error_prop(title, rate, monitors, attenuators=None,
                    norm='monitor'):
    """
    Compare point averaging algorithms on simulated data.

    *title* is the label for the comparison.

    *rate* is the underlying count rate.

    *monitors* is the set of counting intervals, or counting time if
    Plugging in some numbers combining two counts, one for 2000 monitors
    and one for 4000 monitors, we can compute the difference between
    measuring the values separately and measuring them together.  Using
    count rates which are low relative to the monitor yields a relative
    error on the order of 1/10^6:

        Na=7, Ma=2000, Nb=13, Mb=4000, Aa=Ab=1, dAa=dAb=0

    When count rates are high, such as with a direct beam measurement where
    the monitor rate 10% of the detector rate, the relative error is on
    the order of 0.02%::

        Na=20400, Ma=2000, Nb=39500, Mb=4000

    Monitor uncertainty is significant at high count rates.  In the above
    example, the rate uncertainty dr/r for Na/Ma is found to be 2.4% when
    monitor uncertainty is included, but only 0.7% if monitor uncertainty is
    not included in the calculation.  At high Q, where uncertainty count rates
    are much lower than the monitor rate, monitor uncertainty is much less
    important.

    The effect of Poisson versus gaussian averaging is marginal, even for
    regions with extremely low counts, so long as the gaussian average is
    weighted by monitor counts.
    """
    from uncertainties import ufloat
    from uncertainties.unumpy import uarray, nominal_values as uval, \
        std_devs as udev
    from numpy.random import poisson

    #norm='time'
    #norm='monitor'
    time_err = 0
    #time_err = 0.1

    use_attenuators = attenuators is not None
    monitors = np.array(monitors, 'd')
    if norm == 'monitor':
        umonitors = uarray(monitors, np.sqrt(monitors))
    else:
        umonitors = uarray(monitors, time_err)
    if use_attenuators:
        uattenuators = uarray(*list(zip(*attenuators)))
    else:
        uattenuators = uarray(monitors/monitors, 0.*monitors)
    expected = rate*monitors/uval(uattenuators)
    # TODO: simulation is not correct for count by monitor
    counts = poisson(expected)    # Simulated counts
    #counts = expected   # Non-simulated counts
    #ucounts = uarray(counts, np.sqrt(counts + (counts==0)))
    #ucounts = uarray(counts + (counts==0), np.sqrt(counts + (counts==0)))
    ucounts = uarray(counts, np.sqrt(counts))
    incident = ucounts*uattenuators

    print("="*10, title+", rate=%g,"%rate, \
          "median counts=%d"%np.median(counts), "="*10)

    # rate averaged across different counting intervals
    y = incident/umonitors
    y_ave = poisson_average(uval(y), udev(y), norm=norm)
    y_gm = np.sum(monitors*y)/np.sum(monitors)
    y_ave = ufloat(*y_ave)
    y_g = np.mean(y)

    # rate estimated from full counting time

    tin = (np.sum(incident) if use_attenuators
           else ufloat(np.sum(counts), np.sqrt(np.sum(counts))))
    if norm == 'monitor':
        tmon = ufloat(np.sum(monitors), np.sqrt(np.sum(monitors)))
    else:
        tmon = ufloat(np.sum(monitors), len(monitors)*time_err)
    direct = tin/tmon

    #print("monitors:", monitors)
    #print("counts:", counts)
    #print("incident:", incident)
    if use_attenuators:
        print("attenuators:", list(zip(*attenuators))[0])
    def show(label, r, tag=""):
        if r is direct:
            rel = ""
        else:
            rel = (" diff: (%.1f,%.1f)%% "
                   % (100*(r.n-direct.n)/direct.n, 100*(r.s-direct.s)/direct.s))
        print(label, "r:", r, " dr/r: %.2f%%"%(100*r.s/r.n), rel, tag)
    show("Combined", direct)
    show("Poisson ", y_ave)
    show("Gaussian", y_g)
    show("Gaussian", y_gm, "monitor weighted")

    # again without monitor uncertainty
    if 0:
        y2 = incident/monitors
        y2_ave = ufloat(*poisson_average(uval(y2), udev(y2)))
        y2_g = np.mean(y2)
        show("Separate", y2_ave, "no monitor uncertainty")
        show("Gaussian", y2_g, "no monitor uncertainty")

if __name__ == "__main__":
    if 0:
        demo_error_prop("mixed monitor", 10.0, [2000, 2000, 4000])
        demo_error_prop("mixed monitor", 1.0, [2000, 2000, 4000])
        demo_error_prop("mixed monitor", 0.035, [2000, 2000, 4000])
        demo_error_prop("mixed monitor", 0.0035, [2000, 2000, 4000])
        #demo_error_prop("mixed monitor", 0.00035, [2000, 2000, 4000])
        demo_error_prop("same monitor", 10.0, [2000]*700)
        demo_error_prop("same monitor", 1.0, [2000]*700)
        demo_error_prop("same monitor", 0.035, [2000]*700)
        demo_error_prop("same monitor", 0.0035, [2000]*700)
        demo_error_prop("same monitor", 0.00035, [2000]*700)
        demo_error_prop("attenuators", 10.0, [2000, 2000, 4000], [(1, 0), (12, 0.2), (12, 0.2)])
        demo_error_prop("attenuators", 10.0, [2000, 4000], [(12, 0.2), (122, 0.2)])
    else:
        demo_sub()
